"""Methods to deal with Parameter Configuration Space files."""
from pathlib import Path
from sparkle.solver import Solver


def write_configuration_pcs(solver: Solver, config_str: str, tmp_path: Path) -> None:
    """Write configuration to a new PCS file.

    Args:
        solver: Solver object
        config_str: Configuration to write
        tmp_path: Path to place the latest configuration pcs
    """
    # Read optimised configuration and convert to dict
    optimised_configuration_list = config_str.split()

    # Create dictionary
    config_dict = {}
    for i in range(0, len(optimised_configuration_list), 2):
        # Remove dashes and spaces from parameter names, and remove quotes and
        # spaces from parameter values before adding them to the dict
        config_dict[optimised_configuration_list[i].strip(" -")] = (
            optimised_configuration_list[i + 1].strip(" '"))

    # Read existing PCS file and create output content
    pcs_file = solver.get_pcs_file()
    pcs_file_out = []

    with pcs_file.open("r") as infile:
        for line in infile:
            # Copy empty lines
            if not line.strip():
                line_out = line
            # Don't mess with conditional (containing '|') and forbidden (starting
            # with '{') parameter clauses, copy them as is
            elif "|" in line or line.startswith("{"):
                line_out = line
            # Also copy parameters that do not appear in the optimised list
            # (if the first word in the line does not match one of the parameter names
            # in the dict)
            elif line.split()[0] not in config_dict:
                line_out = line
            # Modify default values with optimised values
            else:
                words = line.split("[")
                if len(words) == 2:
                    # Second element is default value + possible tail
                    param_name = line.split()[0]
                    param_val = config_dict[param_name]
                    tail = words[1].split("]")[1]
                    line_out = words[0] + "[" + param_val + "]" + tail
                elif len(words) == 3:
                    # Third element is default value + possible tail
                    param_name = line.split()[0]
                    param_val = config_dict[param_name]
                    tail = words[2].split("]")[1]
                    line_out = words[0] + words[1] + "[" + param_val + "]" + tail
                else:
                    # This does not seem to be a line with a parameter definition, copy
                    # as is
                    line_out = line
            pcs_file_out.append(line_out)

    latest_configuration_pcs_path = tmp_path / "latest_configuration.pcs"

    with latest_configuration_pcs_path.open("w") as outfile:
        for element in pcs_file_out:
            outfile.write(str(element))



#!/usr/bin/env python3
"""Sparkle command to execute a portfolio selector."""

import sys
import argparse
from pathlib import PurePath

from runrunner import Runner

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.support import run_portfolio_selector_help as srpsh
from sparkle.CLI.help import sparkle_logging as sl
from sparkle.platform.settings_objects import Settings, SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.types.objective import PerformanceMeasure
from sparkle.structures import PerformanceDataFrame
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name
from sparkle.instance import InstanceSet


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.InstancePathPositional.names,
                        **ac.InstancePathPositional.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*ac.PerformanceMeasureArgument.names,
                        **ac.PerformanceMeasureArgument.kwargs)

    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    instance_set = resolve_object_name(
        args.instance_path,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, InstanceSet)

    check_for_initialise(
        COMMAND_DEPENDENCIES[CommandName.RUN_SPARKLE_PORTFOLIO_SELECTOR]
    )

    if ac.set_by_user(args, "settings_file"):
        gv.settings().read_settings_ini(
            args.settings_file, SettingState.CMD_LINE
        )  # Do first, so other command line options can override settings from the file
    if ac.set_by_user(args, "performance_measure"):
        gv.settings().set_general_sparkle_objectives(
            args.performance_measure, SettingState.CMD_LINE
        )

    # Compare current settings to latest.ini
    prev_settings = Settings(PurePath("Settings/latest.ini"))
    Settings.check_settings_changes(gv.settings(), prev_settings)

    if gv.settings().get_general_sparkle_objectives()[0].PerformanceMeasure\
            == PerformanceMeasure.QUALITY_ABSOLUTE:
        print("ERROR: The run_sparkle_portfolio_selector command is not yet implemented"
              " for the QUALITY_ABSOLUTE performance measure!")
        sys.exit(-1)
    selector_path = gv.settings().DEFAULT_selection_output / "sparkle_portfolio_selector"
    if not selector_path.exists() or not selector_path.is_file():
        print("ERROR: The portfolio selector could not be found. Please make sure to "
              "first construct a portfolio selector.")
        sys.exit(-1)

    # Multipe
    if instance_set.size > 1:
        test_case_path = gv.settings().DEFAULT_selection_output_test / instance_set.name
        test_case_path.mkdir(parents=True, exist_ok=True)
        # Update latest scenario
        gv.latest_scenario().set_selection_test_case_directory(test_case_path)
        gv.latest_scenario().set_latest_scenario(Scenario.SELECTION)
        # Write used scenario to file
        gv.latest_scenario().write_scenario_ini()
        test_performance_data = PerformanceDataFrame(
            test_case_path / "sparkle_performance_data.csv",
            objectives=gv.settings().get_general_sparkle_objectives())
        run = srpsh.run_portfolio_selector_on_instances(
            instance_set.instance_paths, test_performance_data, selector_path,
            run_on=run_on)
        if run_on == Runner.LOCAL:
            run.wait()
            print("Running Sparkle portfolio selector done!")
        else:
            print("Sparkle portfolio selector is running ...")
    # Single instance
    elif instance_set.size == 1:
        instance = instance_set.instance_paths[0]
        if instance_set.multi_file:
            instance_set._instance_names[0]
        srpsh.portfolio_selector_solve_instance(instance)
        print("Running Sparkle portfolio selector done!")
    else:
        print("Input instance or instance directory error!")

    # Write used settings to file
    gv.settings().write_used_settings()



"""Objects and methods relating to instances for Sparkle."""
from __future__ import annotations
from pathlib import Path
import csv


class InstanceSet:
    """Object representation of a set of instances."""
    instance_csv = "instances.csv"

    def __init__(self: InstanceSet, directory: Path) -> None:
        """Initialise an Instances object from a directory.

        Args:
            directory: Path to the instances directory. If it contains an instance list
                for multi file instances, will read from instance_list.csv. If the path
                is a file, will create an Instance set of size one.
        """
        self.directory: Path = directory
        self.name: str = directory.name
        self.multi_file: bool = False
        self._instance_names: list[str] = []
        self.instance_paths: list[Path] = []

        if self.directory.is_file():
            # Single instance set
            self.instance_paths = [self.directory]
            self._instance_names = [self.directory.name]
            self.directory = self.directory.parent
        elif (self.directory / InstanceSet.instance_csv).exists():
            # Dealing with multiple files per instance
            self.multi_file = True
            # A multi instance file describes per line: InstanceName, File1, File2, ...
            # where each file is present in the self.directory
            for line in csv.reader((self.directory / InstanceSet.instance_csv).open()):
                self._instance_names.append(line[0])
                self.instance_paths.append([(self.directory / f) for f in line[1:]])
        else:
            # Default situation, treat each file in the directory as an instance
            self.instance_paths = [p for p in self.directory.iterdir()]
            self._instance_names = [p.name for p in self.instance_paths]

    @property
    def size(self: InstanceSet) -> int:
        """Returns the number of instances in the set."""
        return len(self.instance_paths)

    @property
    def all_paths(self: InstanceSet) -> list[Path]:
        """Returns all file paths in the instance set as a flat list."""
        if self.multi_file:
            return [p for instance in self.instance_paths for p in instance] + [
                self.directory / InstanceSet.instance_csv]
        return self.instance_paths

    @property
    def get_instance_paths(self: InstanceSet) -> list[Path]:
        """Get processed instance paths for multi-file instances."""
        if self.multi_file:
            return [self.directory / name for name in self._instance_names]
        return self.instance_paths

    def get_path_by_name(self: InstanceSet, name: str) -> Path | list[Path]:
        """Retrieves an instance paths by its name. Returns None upon failure."""
        for idx, instance_name in enumerate(self._instance_names):
            if instance_name == name:
                return self.instance_paths[idx]
        return None



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Module to manage performance data CSV files and common operations on them."""


from __future__ import annotations
from typing import Callable
import operator
from pathlib import Path
import sys
import math
from statistics import mean
import pandas as pd

from sparkle.types.objective import SparkleObjective


class PerformanceDataFrame():
    """Class to manage performance data and common operations on them."""
    missing_value = math.nan
    missing_objective = "DEFAULT:UNKNOWN"
    multi_dim_names = ["Objective", "Instance", "Run"]

    def __init__(self: PerformanceDataFrame,
                 csv_filepath: Path,
                 solvers: list[str] = [],
                 objectives: list[str | SparkleObjective] = None,
                 instances: list[str] = [],
                 n_runs: int = 1,
                 init_df: bool = True) -> None:
        """Initialise a PerformanceDataFrame.

        Consists of:
            - Columns representing the Solvers
            - Rows representing the result by multi-index in order of:
                * Objective (Static, given in constructor or read from file)
                * Instance
                * Runs (Static, given in constructor or read from file)

        Args:
            csv_filepath: If path exists, load from Path.
                Otherwise create new and save to this path.
            solvers: List of solver names to be added into the Dataframe
            objectives: List of SparkleObjectives or objective names. By default None,
                then the objectives will be derived from Sparkle Settings if possible.
            instances: List of instance names to be added into the Dataframe
            n_runs: The number of runs to consider per Solver/Objective/Instance comb.
            init_df: Whether the dataframe should be initialised. Set to false to reduce
                heavy IO loads.
        """
        self.csv_filepath = csv_filepath
        # Runs is a ``static'' dimension
        self.n_runs = n_runs
        self.run_ids = list(range(1, self.n_runs + 1))  # We count runs from 1
        if init_df:
            if self.csv_filepath.exists():
                self.dataframe = pd.read_csv(csv_filepath)
                has_rows = len(self.dataframe.index) > 0
                if (PerformanceDataFrame.multi_dim_names[0] not in self.dataframe.columns
                        or not has_rows):
                    # No objective present, force into column
                    if objectives is None:
                        self.dataframe[PerformanceDataFrame.multi_dim_names[0]] =\
                            PerformanceDataFrame.missing_objective
                    else:  # Constructor is provided with the objectives
                        objectives = [o.name if isinstance(o, SparkleObjective) else o
                                      for o in objectives]
                        self.dataframe[PerformanceDataFrame.multi_dim_names[0]] =\
                            objectives
                if (PerformanceDataFrame.multi_dim_names[2] not in self.dataframe.columns
                        or not has_rows):
                    # No runs column present, force into column
                    self.n_runs = 1
                    self.dataframe[PerformanceDataFrame.multi_dim_names[2]] = self.n_runs
                    self.run_ids = [self.n_runs]
                else:
                    # Runs are present, determine run ids
                    run_label = PerformanceDataFrame.multi_dim_names[2]
                    self.run_ids = self.dataframe[run_label].unique().tolist()
                if PerformanceDataFrame.multi_dim_names[1] not in self.dataframe.columns:
                    # Instances are listed as rows, force into column
                    self.dataframe = self.dataframe.reset_index().rename(
                        columns={"index": PerformanceDataFrame.multi_dim_names[1]})
                # Now we can cast the columns into multi dim
                self.dataframe = self.dataframe.set_index(
                    PerformanceDataFrame.multi_dim_names)
            else:
                # Initialize empty DataFrame
                if objectives is None:
                    objective_names = [PerformanceDataFrame.missing_objective]
                else:
                    objective_names =\
                        [o.name if isinstance(o, SparkleObjective) else o
                         for o in objectives]
                midx = pd.MultiIndex.from_product(
                    [objective_names, instances, self.run_ids],
                    names=PerformanceDataFrame.multi_dim_names)
                self.dataframe = pd.DataFrame(PerformanceDataFrame.missing_value,
                                              index=midx,
                                              columns=solvers)
                self.save_csv()
            # Sort the index to optimize lookup speed
            self.dataframe = self.dataframe.sort_index()

    def verify_objective(self: PerformanceDataFrame,
                         objective: str) -> str:
        """Method to check whether the specified objective is valid.

        Users are allowed to index the dataframe without specifying all dimensions.
        However, when dealing with multiple objectives this is not allowed and this
        is verified here. If we have only one objective this is returned. Otherwise,
        if an objective is specified by the user this is returned.

        Args:
            objective: The objective given by the user
        """
        if objective is None:
            if self.multi_objective:
                print("Error: MO Performance Data, but objective not specified.")
                sys.exit(-1)
            elif self.num_objectives == 1:
                return self.objective_names[0]
            else:
                return PerformanceDataFrame.missing_objective
        return objective

    def verify_run_id(self: PerformanceDataFrame,
                      run_id: int) -> int:
        """Method to check whether run id is valid.

        Similar to verify_objective but here we check the dimensionality of runs.

        Args:
            run_id: the run as specified by the user.
        """
        if run_id is None:
            if self.n_runs > 1:
                print("Error: Multiple run performance data, but run not specified")
                sys.exit(-1)
            else:
                run_id = self.run_ids[0]
        return run_id

    def verify_indexing(self: PerformanceDataFrame,
                        objective: str,
                        run_id: int) -> tuple[str, int]:
        """Method to check whether data indexing is correct.

        Users are allowed to use the Performance Dataframe without the second and
        fourth dimension (Objective and Run respectively) in the case they only
        have one objective or only do one run. This method adjusts the indexing for
        those cases accordingly.

        Args:
            objective: The given objective name
            run_id: The given run index

        Returns:
            A tuple representing the (possibly adjusted) Objective and Run index.
        """
        objective = self.verify_objective(objective)
        run_id = self.verify_run_id(run_id)
        return objective, run_id

    def add_solver(self: PerformanceDataFrame,
                   solver_name: str,
                   initial_value: float | list[float] = None) -> None:
        """Add a new solver to the dataframe. Initializes value to None by default.

        Args:
            solver_name: The name of the solver to be added.
            initial_value: The value assigned for each index of the new solver.
                If not None, must match the index dimension (n_obj * n_inst * n_runs).
        """
        if solver_name in self.dataframe.columns:
            print(f"WARNING: Tried adding already existing solver {solver_name} to "
                  f"Performance DataFrame: {self.csv_filepath}")
            return
        self.dataframe[solver_name] = initial_value

    def add_instance(self: PerformanceDataFrame,
                     instance_name: str,
                     initial_value: float | list[float] = None) -> None:
        """Add and instance to the DataFrame."""
        if self.dataframe.index.size == 0 or self.dataframe.columns.size == 0:
            # First instance or no Solvers yet
            solvers = self.dataframe.columns.to_list()
            instances = self.dataframe.index.levels[1].to_list() + [instance_name]
            midx = pd.MultiIndex.from_product(
                [self.objective_names, instances, self.run_ids],
                names=PerformanceDataFrame.multi_dim_names)
            self.dataframe = pd.DataFrame(initial_value, index=midx, columns=solvers)
        else:
            if instance_name in self.dataframe.index.levels[1]:
                print(f"WARNING: Tried adding already existing instance {instance_name} "
                      f"to Performance DataFrame: {self.csv_filepath}")
                return
            # Create the missing indices
            levels = [self.dataframe.index.levels[0].tolist(),
                      [instance_name],
                      self.dataframe.index.levels[2].tolist()]
            emidx = pd.MultiIndex(levels, names=PerformanceDataFrame.multi_dim_names)
            # Create the missing column values
            edf = pd.DataFrame(PerformanceDataFrame.missing_value,
                               index=emidx,
                               columns=self.dataframe.columns)
            # Concatenate the original and new dataframe together
            self.dataframe = pd.concat([self.dataframe, edf])

    # Can we make this handle a sequence of inputs instead of just 1?
    def set_value(self: PerformanceDataFrame,
                  value: float,
                  solver: str,
                  instance: str,
                  objective: str = None,
                  run: int = None) -> None:
        """Setter method to assign a value to the Dataframe.

        Args:
            value: Float value to be assigned.
            solver: The solver that produced the value.
            instance: The instance that the value was produced on.
            objective: The objective for which the result was produced.
                Optional in case of using single objective.
            run: The run index for which the result was produced.
                Optional in case of doing single run results.
        """
        objective, run = self.verify_indexing(objective, run)
        self.dataframe.at[(objective, instance, run), solver] = value

    def remove_solver(self: PerformanceDataFrame, solver_name: str) -> None:
        """Drop a solver from the Dataframe."""
        self.dataframe.drop(solver_name, axis=1, inplace=True)

    def remove_instance(self: PerformanceDataFrame, instance_name: str) -> None:
        """Drop an instance from the Dataframe."""
        self.dataframe.drop(instance_name, axis=0, level="Instance", inplace=True)

    def reset_value(self: PerformanceDataFrame,
                    solver: str,
                    instance: str,
                    objective: str = None,
                    run: int = None) -> None:
        """Reset a value in the dataframe."""
        self.set_value(PerformanceDataFrame.missing_value,
                       solver, instance, objective, run)

    def get_value(self: PerformanceDataFrame,
                  solver: str,
                  instance: str,
                  objective: str = None,
                  run: int = None) -> float:
        """Index a value of the DataFrame and return it."""
        objective, run = self.verify_indexing(objective, run)
        return self.dataframe.loc[(objective, instance, run), solver]

    @property
    def num_objectives(self: PerformanceDataFrame) -> int:
        """Retrieve the number of objectives in the DataFrame."""
        return self.dataframe.index.levels[0].size

    @property
    def num_instances(self: PerformanceDataFrame) -> int:
        """Return the number of instances."""
        return self.dataframe.index.levels[1].size

    @property
    def num_runs(self: PerformanceDataFrame) -> int:
        """Return the number of runs."""
        return self.dataframe.index.levels[2].size

    @property
    def num_solvers(self: PerformanceDataFrame) -> int:
        """Return the number of solvers."""
        return self.dataframe.columns.size

    @property
    def multi_objective(self: PerformanceDataFrame) -> bool:
        """Return whether the dataframe represent MO or not."""
        return self.num_objectives > 1

    @property
    def solvers(self: PerformanceDataFrame) -> list[str]:
        """Return the solver present as a list of strings."""
        return self.dataframe.columns.tolist()

    @property
    def objective_names(self: PerformanceDataFrame) -> list[str]:
        """Return the objective names as a list of strings."""
        if self.num_objectives == 0:
            return [PerformanceDataFrame.missing_objective]
        return self.dataframe.index.levels[0].tolist()

    @property
    def instances(self: PerformanceDataFrame) -> list[str]:
        """Return the instances as a Pandas Index object."""
        return self.dataframe.index.levels[1].tolist()

    def penalise(self: PerformanceDataFrame,
                 threshold: float,
                 penalty: float,
                 objective: str = None,
                 lower_bound: bool = False) -> None:
        """Penalises the DataFrame values if crossing threshold by specified penalty.

        Directly updates the DataFrame object held by this class.

        Args:
            threshold: The threshold of performances to be met
            penalty: The values assigned for out of bounds performances
            objective: The objective that should be penalised.
            lower_bound: Whether the threshold is a lower_bound. By default,
                the threshold is treated as an upperbound for performance values.
        """
        objective = self.verify_objective(objective)
        comparison_op = operator.lt if lower_bound else operator.gt
        self.dataframe[comparison_op(self.dataframe.loc[(objective), :],
                                     threshold)] = penalty

    def get_job_list(self: PerformanceDataFrame, rerun: bool = False) \
            -> list[tuple[str, str]]:
        """Return a list of performance computation jobs there are to be done.

        Get a list of tuple[instance, solver] to run from the performance data
        csv file. If rerun is False (default), get only the tuples that don't have a
        value in the table, else (True) get all the tuples.

        Args:
            rerun: Boolean indicating if we want to rerun all jobs
        """
        df = self.dataframe.stack(future_stack=True)
        if not rerun:
            df = df[df.isnull()]
        if not self.multi_objective:
            df.index = df.index.droplevel(["Objective"])
        if self.n_runs == 1:
            df.index = df.index.droplevel(["Run"])
        return df.index.tolist()

    def has_missing_performance(self: PerformanceDataFrame) -> bool:
        """Returns True if there are any missing values in the dataframe."""
        return self.dataframe.isnull().any().any()

    def remaining_jobs(self: PerformanceDataFrame) -> dict[str, list[str]]:
        """Return a dictionary for empty values per instance and solver combination."""
        remaining_jobs = {}
        null_df = self.dataframe.isnull()
        for row in self.dataframe.index:
            instance = row[1]
            for solver in self.dataframe.columns:
                if null_df.at[row, solver]:
                    if instance not in remaining_jobs:
                        remaining_jobs[instance] = set([solver])
                    else:
                        remaining_jobs[instance].add(solver)
        return remaining_jobs

    def get_best_performance_per_instance(
            self: PerformanceDataFrame,
            objective: str = None,
            run_agg: Callable = None,
            best: Callable = pd.DataFrame.min) -> list[float]:
        """Return a list with the best performance per instance.

        Allows user to specify how to aggregate runs together.
        Must be pandas accepted callable.

        Args:
            objective: The objective over which we want the ``best''
            run_agg: The method defining how runs are combined per Instance/solver
            best: Callable, replace with pd.DataFrame.max to maximise

        Returns:
            A list of floats representing the best performance per instance
        """
        objective = self.verify_objective(objective)
        if run_agg is None:
            return best(self.dataframe.loc[(objective), :], axis=1).to_list()
        else:
            # Select the objective, group by runs, apply the aggregator and select max
            agg_runs = self.dataframe.loc[(objective), :].groupby(level=0).run_agg()
            return best(agg_runs, axis=1).to_list()

    def best_performance_instance(
            self: PerformanceDataFrame,
            instance: str,
            minimise: bool,
            objective: str = None,
            exclude_solvers: list[str] = [],
            run_aggregator: Callable = mean) -> float:
        """Return the best performance for a specific instance.

        Args:
            instance: The instance in question
            minimise: Whether we should minimise or maximise the score
            objective: The objective for which we calculate the best performance
            capvalue: The minimum/maximum scoring value the VBS is allowed to have
            run_aggregator: How we aggregate multiple runs for an instance-solver
                combination. Only relevant for multi-runs.
            exclude_solvers: List of solvers to exclude in the calculation.

        Returns:
            The virtual best solver performance for this instance.
        """
        objective = self.verify_objective(objective)
        best_score = None
        compare = operator.lt if minimise else operator.gt
        for solver in self.solvers:
            if solver in exclude_solvers:
                continue
            if isinstance(instance, str):
                runs = self.dataframe.loc[(objective, instance), solver]
                score_solver = run_aggregator(runs)
            else:
                score_solver = float(self.dataframe.loc[instance, solver])
            if best_score is None or compare(score_solver, best_score):
                best_score = score_solver

        if best_score is None:
            print("WARNING: PerformanceDataFrame could not calculate best performance "
                  f"for instance {instance}.")
            return PerformanceDataFrame.missing_value

        return best_score

    def best_performance(
            self: PerformanceDataFrame,
            aggregation_function: Callable[[list[float]], float],
            minimise: bool,
            exclude_solvers: list[str] = [],
            objective: str = None) -> float:
        """Return the overall best performance of the portfolio.

        Args:
            aggregation_function: The method of combining all VBS scores together
            minimise: Whether the scores are minimised or not
            capvalue_list: List of capvalue per instance
            penalty_list: List of penalty per instance
            exclude_solvers: List of solvers to exclude in the calculation.
                Defaults to none.

        Returns:
            The aggregated best performance of the portfolio over all instances.
        """
        objective = self.verify_objective(objective)
        instance_best = []
        for instance in self.instances:
            best_score = self.best_performance_instance(
                instance, minimise, objective, exclude_solvers)
            instance_best.append(best_score)

        return aggregation_function(instance_best)

    def marginal_contribution(self: PerformanceDataFrame,
                              aggregation_function: Callable[[list[float]], float],
                              minimise: bool,
                              pre_selection: list[str] = None,
                              objective: str = None,
                              ) -> list[float]:
        """Return the marginal contribution of the solvers on the instances.

        Args:
            aggregation_function: Function for combining scores over instances together.
            minimise: Whether we should minimise or maximise the score of the objective.
            capvalue_list: List of capvalue per instance
            penalty_list: List of penalty per instance
            objective: The objective for which we calculate the marginal contribution.

        Returns:
            The marginal contribution of each solver.
        """
        output = []
        objective = self.verify_objective(objective)
        best_performance = self.best_performance(
            aggregation_function, minimise, objective=objective)
        for solver in self.solvers:
            # By calculating the best performance excluding this Solver,
            # we can determine its relative impact on the portfolio.
            missing_solver_best = self.best_performance(
                aggregation_function,
                minimise,
                exclude_solvers=[solver],
                objective=objective)
            # Now we need to see how much the portfolio's best performance
            # decreases without this solver.
            marginal_contribution = missing_solver_best / best_performance
            if missing_solver_best == best_performance:
                # No change, no contribution
                marginal_contribution = 0.0
            output.append((solver, marginal_contribution, missing_solver_best))
        return output

    def get_dict_vbs_penalty_time_on_each_instance(
            self: PerformanceDataFrame,
            penalised_time: int,
            objective: str = None,
            run_id: int = None) -> dict:
        """Return a dictionary of penalised runtimes and instances for the VBS."""
        objective = self.verify_objective(objective)
        instance_penalized_runtimes = {}
        for instance in self.dataframe.index.levels[1]:
            if run_id is None:
                runtime = self.dataframe.loc[(objective, instance), :].min(axis=None)
            else:
                runtime =\
                    self.dataframe.loc[(objective, instance, run_id), :].min(axis=None)
            instance_penalized_runtimes[instance] = min(penalised_time, runtime)

        return instance_penalized_runtimes

    def calc_vbs_penalty_time(self: PerformanceDataFrame,
                              cutoff_time: int = None,
                              penalty: int = None,
                              objective: str = None) -> float:
        """Return the penalised performance of the VBS."""
        objective = self.verify_objective(objective)
        if cutoff_time is not None and penalty is not None:
            self.penalise(cutoff_time, penalty)
        # Calculate the minimum for the selected objective per instance
        min_instance_df = self.dataframe.loc(axis=0)[objective, :, :].min(axis=1)
        # Return average
        return min_instance_df.sum() / self.dataframe.index.size

    def get_solver_penalty_time_ranking(self: PerformanceDataFrame,
                                        cutoff_time: int = None,
                                        penalty: int = None,
                                        objective: str = None,
                                        ) -> list[list[float]]:
        """Return a list with solvers ranked by penalised runtime."""
        objective = self.verify_objective(objective)
        if cutoff_time is not None and penalty is not None:
            self.penalise(cutoff_time, penalty, objective)
        solver_penalty_time_ranking = []
        num_instances = self.dataframe.index.size
        sub_df = self.dataframe.loc(axis=0)[objective, :, :]
        for solver in self.dataframe.columns:
            average_time = sub_df[solver].sum() / num_instances
            solver_penalty_time_ranking.append([solver, average_time])

        # Sort the list by second value (the penalised run time)
        solver_penalty_time_ranking.sort(
            key=lambda this_penalty_time: this_penalty_time[1])

        return solver_penalty_time_ranking

    def save_csv(self: PerformanceDataFrame, csv_filepath: Path = None) -> None:
        """Write a CSV to the given path.

        Args:
            csv_filepath: String path to the csv file. Defaults to self.csv_filepath.
        """
        csv_filepath = self.csv_filepath if csv_filepath is None else csv_filepath
        self.dataframe.to_csv(csv_filepath)

    def clean_csv(self: PerformanceDataFrame) -> None:
        """Set all values in Performance Data to None."""
        self.dataframe[:] = PerformanceDataFrame.missing_value
        self.save_csv()

    def copy(self: PerformanceDataFrame,
             csv_filepath: Path = None) -> PerformanceDataFrame:
        """Create a copy of this object.

        Args:
            csv_filepath: The new filepath to use for saving the object to.
                Warning: If the original path is used, it could lead to dataloss!
        """
        csv_filepath = self.csv_filepath if csv_filepath is None else csv_filepath
        pd_copy = PerformanceDataFrame(self.csv_filepath, init_df=False)
        pd_copy.dataframe = self.dataframe.copy()
        pd_copy.csv_filepath = csv_filepath
        return pd_copy

    def to_autofolio(self: PerformanceDataFrame) -> Path:
        """Port the data to a format acceptable for AutoFolio."""
        if self.multi_objective or self.n_runs > 1:
            print(f"ERROR: Currently no porting available for {self.csv_filepath} "
                  "to Autofolio due to multi objective or number of runs.")
            return
        autofolio_df = self.dataframe.copy()
        autofolio_df.index = autofolio_df.index.droplevel(["Objective", "Run"])
        path = self.csv_filepath.parent / f"autofolio_{self.csv_filepath.name}"
        autofolio_df.to_csv(path)
        return path



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper classes/method for LaTeX and bibTeX."""
from shutil import which
from pathlib import Path
import subprocess
from enum import Enum


class ReportType(str, Enum):
    """enum for separating different types of reports."""
    ALGORITHM_SELECTION = "algorithm_selection"
    ALGORITHM_CONFIGURATION = "algorithm_configuration"
    PARALLEL_PORTFOLIO = "parallel_portfolio"


def check_tex_commands_exist(latex_directory_path: Path) -> None:
    """Raise an exception if one of the latex commands is not present."""
    if which("bibtex") is None or which("pdflatex") is None:
        raise Exception("Error: It seems like latex is not available on your system.\n"
                        "You can install latex and run the command again, "
                        f"or copy the source files in {latex_directory_path} on your "
                        "local machine to generate the report.")


def compile_pdf(latex_files_path: Path, latex_report_filename: Path) -> Path:
    """Compile the given latex files to a PDF.

    Args:
        latex_files_path: Path to the directory with source files
            where the report will be generated.
        latex_report_filename: Name of the output files.

    Returns:
        Path to the newly generated report in PDF format.
    """
    pdf_process = subprocess.run(["pdflatex", "-interaction=nonstopmode",
                                 f"{latex_report_filename}.tex"],
                                 cwd=latex_files_path, capture_output=True)

    if pdf_process.returncode != 0:
        print(f"[{pdf_process.returncode}] ERROR generating with PDFLatex command:\n"
              f"{pdf_process.stdout.decode()}\n {pdf_process.stderr.decode()}\n")

    bibtex_process = subprocess.run(["bibtex", f"{latex_report_filename}.aux"],
                                    cwd=latex_files_path, capture_output=True)

    if bibtex_process.returncode != 0:
        print("ERROR whilst generating with Bibtex command:"
              f"{bibtex_process.stdout} {bibtex_process.stderr}")

    # TODO: Fix compilation for references
    # (~\ref[] yields [?] in pdf, re-running command fixes it)
    # We have to re-run the same pdf command to take in the updates bib files from bibtex
    # But Bibtex cannot function without .aux file produced by pdflatex. Hence run twice.
    pdf_process = subprocess.run(["pdflatex", "-interaction=nonstopmode",
                                 f"{latex_report_filename}.tex"],
                                 cwd=latex_files_path, capture_output=True)

    return Path(latex_files_path / latex_report_filename).with_suffix(".pdf")



"""Test functionalities related to the compute marginal contribution help module."""

from __future__ import annotations
from unittest import TestCase
from pathlib import Path

from sparkle.CLI.support import compute_marginal_contribution_help as scmch
from sparkle.structures import FeatureDataFrame, PerformanceDataFrame

from unittest.mock import patch
from unittest.mock import MagicMock


class TestMarginalContribution(TestCase):
    """Tests function of Marginal Contribution help."""

    def test_compute_perfect_selector_marginal_contribution(self: TestCase) -> None:
        """Test for method compute_perfect_selector_marginal_contribution."""
        performance_data = PerformanceDataFrame(
            Path("tests/CLI/test_files/Performance_Data/"
                 "test_construct_sparkle_portfolio_selector.csv"))
        result = [("Solvers/CSCCSat", 1.7980089765503102, 33117.589),
                  ("Solvers/MiniSAT", 0.0, 18419.034293999997)]
        output = scmch.compute_perfect_selector_marginal_contribution(
            aggregation_function=sum,
            minimise=True,
            performance_data=performance_data
        )
        self.assertListEqual(output, result)

    @patch("sparkle.CLI.support.compute_marginal_contribution_help."
           "compute_actual_performance_for_instance")
    def test_compute_actual_selector_performance(
            self: TestCase,
            patch_perf_for_instance: MagicMock) -> None:
        """Test for method compute_actual_selector_performance."""
        pth = "tests/CLI/test_files/Sparkle_Portfolio_Selector/"\
              "sparkle_portfolio_selector"
        perf_path = Path("tests/CLI/test_files/Performance_Data/"
                         "test_construct_sparkle_portfolio_selector.csv")
        feature_csv_path = Path("tests/CLI/test_files/Feature_Data/"
                                "test_construct_sparkle_portfolio_selector.csv")

        result = 526.805294
        patch_perf_for_instance.side_effect = [(61.0, False), (28.1747, True),
                                               (61.0, False), (9.98625, True),
                                               (0.107158, True), (61.0, False),
                                               (0.537186, True), (61.0, False),
                                               (61.0, False), (61.0, False),
                                               (61.0, False), (61.0, False)]
        performance_df = PerformanceDataFrame(perf_path)
        feature_df = FeatureDataFrame(feature_csv_path)
        output = scmch.compute_actual_selector_performance(pth,
                                                           performance_df,
                                                           feature_df,
                                                           True,
                                                           sum,
                                                           None)

        assert output == result

    def test_compute_actual_performance_for_instance(self: TestCase) -> None:
        """Test for method compute_actual_performance_for_instance."""
        # TODO: This method is currently not touched by the .sh test. Think of a test.
        pass

    def test_compute_actual_selector_marginal_contribution(self: TestCase
                                                           ) -> None:
        """Test for method compute_actual_selector_marginal_contribution."""
        # TODO: Write test
        pass

    def test_print_rank_list(self: TestCase) -> None:
        """Test for method print_rank_list. Could be irrelevant."""
        # TODO: Write test
        pass

    def test_compute_actual(self: TestCase) -> None:
        """Test for method compute_actual. Could be irrelevant."""
        # TODO: Write test
        pass

    def test_compute_marginal_contribution(self: TestCase) -> None:
        """Test for method compute_marginal_contribution. Could be irrelevant."""
        # TODO: Write test
        pass



#!/usr/bin/env python3
"""Sparkle command to wait for one or more other commands to complete execution."""

import sys
import argparse

from sparkle.CLI.help import sparkle_logging as sl
from sparkle.CLI.support import sparkle_job_help as sjh
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import global_variables as gv


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments.

    Returns:
      The argument parser.
    """
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument(*ac.JobIdArgument.names,
                       **ac.JobIdArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    if args.job_id is not None:
        sjh.wait_for_job(args.job_id, path=gv.settings().DEFAULT_tmp_output)
    else:
        sjh.wait_for_all_jobs(path=gv.settings().DEFAULT_tmp_output)



#!/usr/bin/env python3
"""Sparkle command to remove temporary files.

Only removes files not affecting the sparkle state.
"""

import sys
import argparse

from sparkle.platform import file_help as sfh
from sparkle.CLI.help import sparkle_logging as sl


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()

    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    print("Cleaning temporary files ...")
    sfh.remove_temporary_files()
    sfh.create_temporary_directories()
    print("Temporary files cleaned!")



"""File to handle a solver and its directories."""

from __future__ import annotations
import sys
from typing import Any
import shlex
import ast
from pathlib import Path
import subprocess
from sparkle.tools import runsolver_parsing
import pcsparser
from sparkle.types import SparkleCallable, SolverStatus


class Solver(SparkleCallable):
    """Class to handle a solver and its directories."""
    meta_data = "solver_meta.txt"
    wrapper = "sparkle_solver_wrapper.py"

    def __init__(self: Solver,
                 directory: Path,
                 raw_output_directory: Path = None,
                 runsolver_exec: Path = None,
                 deterministic: bool = None) -> None:
        """Initialize solver.

        Args:
            directory: Directory of the solver.
            raw_output_directory: Directory where solver will write its raw output.
                Defaults to directory / tmp
            runsolver_exec: Path to the runsolver executable.
                By default, runsolver in directory.
            deterministic: Bool indicating determinism of the algorithm.
                Defaults to False.
        """
        super().__init__(directory, runsolver_exec, raw_output_directory)
        self.deterministic = deterministic
        self.meta_data_file = self.directory / Solver.meta_data

        if self.raw_output_directory is None:
            self.raw_output_directory = self.directory / "tmp"
            self.raw_output_directory.mkdir(exist_ok=True)
        if self.runsolver_exec is None:
            self.runsolver_exec = self.directory / "runsolver"
        if not self.meta_data_file.exists():
            self.meta_data_file = None
        if self.deterministic is None:
            if self.meta_data_file is not None:
                # Read the parameter from file
                meta_dict = ast.literal_eval(self.meta_data_file.open().read())
                self.deterministic = meta_dict["deterministic"]
            else:
                self.deterministic = False

    def _get_pcs_file(self: Solver) -> Path | bool:
        """Get path of the parameter file.

        Returns:
            Path to the parameter file or False if the parameter file does not exist.
        """
        pcs_files = [p for p in self.directory.iterdir() if p.suffix == ".pcs"]
        if len(pcs_files) != 1:
            # We only consider one PCS file per solver
            return False
        return pcs_files[0]

    def get_pcs_file(self: Solver) -> Path:
        """Get path of the parameter file.

        Returns:
            Path to the parameter file. None if it can not be resolved.
        """
        if not (file_path := self._get_pcs_file()):
            return None
        return file_path

    def read_pcs_file(self: Solver) -> bool:
        """Checks if the pcs file can be read."""
        pcs_file = self._get_pcs_file()
        try:
            parser = pcsparser.PCSParser()
            parser.load(str(pcs_file), convention="smac")
            return True
        except SyntaxError:
            pass
        return False

    def get_pcs(self: Solver) -> dict[str, tuple[str, str, str]]:
        """Get the parameter content of the PCS file."""
        if not (pcs_file := self.get_pcs_file()):
            return None
        parser = pcsparser.PCSParser()
        parser.load(str(pcs_file), convention="smac")
        return [p for p in parser.pcs.params if p["type"] == "parameter"]

    def build_cmd(self: Solver,
                  instance: str | list[str] = None,
                  configuration: dict = None,
                  runsolver_configuration: list[str] = None) -> list[str]:
        """Build the solver call on an instance with a configuration."""
        if isinstance(configuration, str):
            configuration = Solver.config_str_to_dict(configuration)
        if configuration is None:
            configuration = {}
        # Ensure configuration contains required entries for each wrapper
        if "instance" not in configuration:
            configuration["instance"] = instance
        if "solver_dir" not in configuration:
            configuration["solver_dir"] = str(self.directory.absolute())
        if "specifics" not in configuration:
            configuration["specifics"] = ""
        if "run_length" not in configuration:
            configuration["run_length"] = ""
        if "cutoff_time" not in configuration:
            configuration["cutoff_time"] = sys.maxsize
        # Ensure stringification of dictionary will go correctly
        configuration = {key: str(configuration[key]) for key in configuration}
        # Ensure stringifcation of cmd call will go correctly
        solver_cmd = []
        if runsolver_configuration is not None:
            # Ensure stringification of runsolver configuration is done correctly
            runsolver_configuration = [str(runsolver_config) for runsolver_config in
                                       runsolver_configuration]
            # We wrap the solver call in the runsolver executable, by placing it in front
            solver_cmd += [str(self.runsolver_exec.absolute())] + runsolver_configuration
        solver_cmd += [str((self.directory / Solver.wrapper).absolute()),
                       str(configuration)]
        return solver_cmd

    def run(self: Solver, instance: str | list[str],
            configuration: dict = None,
            runsolver_configuration: list[str] = None,
            cwd: Path = None) -> dict[str, str]:
        """Run the solver on an instance with a certain configuration.

        Args:
            instance: The instance to run the solver on, list in case of multi-file
            configuration: The solver configuration to use. Can be empty.
            runsolver_configuration: The runsolver configuration to wrap the solver
                with. If None (default), runsolver will not be used.
            cwd: Path where to execute. Defaults to self.raw_output_directory.

        Returns:
            Solver output dict possibly with runsolver values.
        """
        if cwd is None:
            cwd = self.raw_output_directory
        solver_cmd = self.build_cmd(instance,
                                    configuration,
                                    runsolver_configuration)
        process = subprocess.run(solver_cmd,
                                 cwd=cwd,
                                 capture_output=True)

        # Subprocess resulted in error
        if process.returncode != 0:
            print(f"WARNING: Solver {self.name} execution seems to have failed!\n"
                  f"The used command was: {solver_cmd}\n The error yielded was: \n"
                  f"\t-stdout: '{process.stdout.decode()}'\n"
                  f"\t-stderr: '{process.stderr.decode()}'\n")
            return {"status": SolverStatus.ERROR, }

        return Solver.parse_solver_output(process.stdout.decode(),
                                          runsolver_configuration,
                                          cwd)

    @staticmethod
    def config_str_to_dict(config_str: str) -> dict[str, str]:
        """Parse a configuration string to a dictionary."""
        # First we filter the configuration of unwanted characters
        config_str = config_str.strip().replace("-", "")
        if config_str == "" or config_str == r"{}":
            return {}
        # Then we split the string by spaces, but conserve substrings
        config_list = shlex.split(config_str)
        config_dict = {}
        for index in range(0, len(config_list), 2):
            # As the value will already be a string object, no quotes are allowed in it
            value = config_list[index + 1].strip('"').strip("'")
            config_dict[config_list[index]] = value
        return config_dict

    @staticmethod
    def parse_solver_output(solver_output: str,
                            runsolver_configuration: list[str] = None,
                            cwd: Path = None) -> dict[str, Any]:
        """Parse the output of the solver.

        Args:
            solver_output: The output of the solver run which needs to be parsed
            runsolver_configuration: The runsolver configuration to wrap the solver
                with. If runsolver was not used this should be None.
            cwd: Path where to execute. Defaults to self.raw_output_directory.

        Returns:
            Dictionary representing the parsed solver output
        """
        if runsolver_configuration is not None:
            parsed_output = runsolver_parsing.get_solver_output(runsolver_configuration,
                                                                solver_output,
                                                                cwd)
        else:
            parsed_output = ast.literal_eval(solver_output)

        # cast status attribute from str to Enum
        parsed_output["status"] = SolverStatus(parsed_output["status"])

        return parsed_output



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for the execution of a portfolio selector."""
import sys
from filelock import FileLock, Timeout
from pathlib import Path

import runrunner as rrr
from runrunner.base import Runner, Run

from sparkle.solver import Extractor, Solver
from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame
from sparkle.CLI.support import run_solvers_help as srs

from sparkle.platform import CommandName


# Only called in call_sparkle_portfolio_selector_solve_instance
def call_solver_solve_instance_within_cutoff(
        solver: Solver,
        instance: Path,
        cutoff_time: int,
        performance_data: PerformanceDataFrame = None) -> bool:
    """Call the Sparkle portfolio selector to solve a single instance with a cutoff.

    Args:
        solver: The solver to run on the instance
        instance: The path to the instance
        cutoff_time: The cutoff time for the solver
        performance_data: The dataframe to store the results in

    Returns:
        Whether the instance was solved by the solver
    """
    _, _, cpu_time_penal, _, status, raw_result_path =\
        srs.run_solver_on_instance_and_process_results(solver,
                                                       instance,
                                                       cutoff_time,
                                                       gv.get_seed())
    flag_solved = False
    if status == "SUCCESS" or status == "SAT" or status == "UNSAT":
        flag_solved = True

    if performance_data is not None:
        solver_name = "sparkle_portfolio_selector"
        print(f"Trying to write: {cpu_time_penal}, {solver_name}, {instance}")
        try:
            # Creating a seperate locked file for writing
            lock = FileLock(f"{performance_data.csv_filepath}.lock")
            with lock.acquire(timeout=60):
                # Reload the dataframe to latest version
                performance_data = PerformanceDataFrame(performance_data.csv_filepath)
                performance_data.set_value(cpu_time_penal, solver_name, str(instance))
                performance_data.save_csv()
            lock.release()
        except Timeout:
            print(f"ERROR: Cannot acquire File Lock on {performance_data}.")
    else:
        if flag_solved:
            print(f"Instance solved by solver {solver.name}")
        else:
            print(f"Solver {solver.name} failed to solve the instance with status "
                  f"{status}")
    raw_result_path.unlink(missing_ok=True)
    return flag_solved


# Only called in portfolio_core and run_sparkle_portfolio_selector
def portfolio_selector_solve_instance(
        instance: Path,
        performance_data: PerformanceDataFrame = None) -> None:
    """Call the Sparkle portfolio selector to solve a single instance.

    Args:
        instance: Path to the instance to run on
        performance_data: path to the performance data
    """
    print(f"Running portfolio selector on solving instance {instance} ...")

    cutoff_extractor = gv.settings().get_general_extractor_cutoff_time()
    print(f"Computing features of instance {instance} ...")
    feature_vector = []
    extractor_paths = [p for p in gv.settings().DEFAULT_extractor_dir.iterdir()]
    if len(extractor_paths) == 0:
        print("ERROR: No feature extractor added to Sparkle.")
        sys.exit(-1)
    for extractor_path in extractor_paths:
        extractor = Extractor(Path(extractor_path),
                              gv.settings().DEFAULT_runsolver_exec,
                              gv.settings().DEFAULT_tmp_output)
        # We create a watch log to filter out runsolver output
        runsolver_watch_path = gv.settings().DEFAULT_tmp_output /\
            f"{extractor_path.name}_{instance}.wlog"
        features = extractor.run(instance,
                                 runsolver_args=["--cpu-limit", str(cutoff_extractor),
                                                 "-w", runsolver_watch_path])
        for _, _, value in features:
            feature_vector.append(value)
        runsolver_watch_path.unlink(missing_ok=True)
    print(f"Sparkle computing features of instance {instance} done!")

    print("Sparkle portfolio selector predicting ...")
    selector = gv.settings().get_general_sparkle_selector()
    predict_schedule = selector.run(gv.settings().DEFAULT_algorithm_selector_path,
                                    feature_vector)

    if predict_schedule is None:
        # Selector Failed to produce prediction
        sys.exit(-1)
    print("Predicting done!")
    for solver, cutoff_time in predict_schedule:
        solver = Solver(Path(solver))
        print(f"Calling solver {solver.name} with time budget {cutoff_time} ...")
        flag_solved = call_solver_solve_instance_within_cutoff(
            solver, instance, cutoff_time, performance_data)
        print(f"Calling solver {solver.name} done!")

        if flag_solved:
            return
        else:
            print("The instance is not solved in this call")


# Only called in run_sparkle_portfolio_selector
def run_portfolio_selector_on_instances(
        instances: list[Path],
        performance_data: PerformanceDataFrame,
        portfolio_selector: Path,
        run_on: Runner = Runner.SLURM) -> Run:
    """Call the Sparkle portfolio selector to solve all instances in a directory.

    Args:
        instances: The paths to the instances.
        performance_data: The dataframe to store the results in.
        portfolio_selector: Path to the selector.
        run_on: Whether to run with Slurm or Local.

    Returns:
        RunRunner Run object regarding the selector call.
    """
    for instance_path in instances:
        performance_data.add_instance(str(instance_path))

    performance_data.add_solver(portfolio_selector.name)

    performance_data.save_csv()

    # TODO: Instead of using run_sparkle_portfolio_core.py, we should do here:
    # 1. Get feature data for every feature extractor
    # 2. Use autofolio call to predict the solver per instance with
    #    a dependency on run object of 1
    # 3. Run the solver and place the results in the performance dataframe w dependency
    #    (Difficult, which solver is determined by 2)
    run_core = Path(__file__).parent.resolve() / "core" / "run_sparkle_portfolio_core.py"
    cmd_list = [f"python {run_core} "
                f"--performance-data-csv {performance_data.csv_filepath} "
                f"--instance {instance_path}" for instance_path in instances]
    run = rrr.add_to_queue(
        runner=run_on,
        cmd=cmd_list,
        name=CommandName.RUN_SPARKLE_PORTFOLIO_SELECTOR,
        base_dir=gv.settings().DEFAULT_tmp_output,
        sbatch_options=gv.settings().get_slurm_extra_options(as_args=True),
        srun_options=["-N1", "-n1"])

    return run



#!/usr/bin/env python3
"""Sparkle command for the computation of the marginal contributions."""

import sys
import argparse

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.support import compute_marginal_contribution_help as scmch
from sparkle.CLI.help import sparkle_logging as sl
from sparkle.platform.settings_objects import SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as apc


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(*apc.PerfectArgument.names,
                       **apc.PerfectArgument.kwargs)
    group.add_argument(*apc.ActualArgument.names,
                       **apc.ActualArgument.kwargs)
    parser.add_argument(*apc.RecomputeMarginalContributionArgument.names,
                        **apc.RecomputeMarginalContributionArgument.kwargs)
    parser.add_argument(*apc.SelectorTimeoutArgument.names,
                        **apc.SelectorTimeoutArgument.kwargs)
    parser.add_argument(*apc.PerformanceMeasureArgument.names,
                        **apc.PerformanceMeasureArgument.kwargs)
    parser.add_argument(*apc.SettingsFileArgument.names,
                        **apc.SettingsFileArgument.kwargs)

    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    check_for_initialise(
        COMMAND_DEPENDENCIES[CommandName.COMPUTE_MARGINAL_CONTRIBUTION]
    )

    print("[Deprecated] command, functionality is called automatically by other commands"
          " when needed.")
    if ac.set_by_user(args, "settings_file"):
        gv.settings().read_settings_ini(
            args.settings_file, SettingState.CMD_LINE
        )  # Do first, so other command line options can override settings from the file
    if ac.set_by_user(args, "performance_measure"):
        gv.settings().set_general_sparkle_objectives(
            args.performance_measure, SettingState.CMD_LINE
        )

    scmch.compute_marginal_contribution(
        args.perfect, args.actual, args.recompute, args.selector_timeout
    )

    # Write used settings to file
    gv.settings().write_used_settings()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for marginal contribution computation."""
from __future__ import annotations
import sys
import ast
import operator
from pathlib import Path
from typing import Callable
from statistics import mean
import tabulate

from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.CLI.construct_sparkle_portfolio_selector\
    import construct_sparkle_portfolio_selector
from sparkle.types.objective import PerformanceMeasure


def compute_perfect_selector_marginal_contribution(
        performance_data: PerformanceDataFrame,
        aggregation_function: Callable[[list[float]], float] = mean,
        minimise: bool = True) -> list[tuple[str, float]]:
    """Return the marginal contributions of solvers for the VBS.

    Args:
      performance_data: Performance DataFrame to compute the contribution for.
      aggregation_function: function to aggregate the per instance scores
      minimise: flag indicating if scores should be minimised or maximised

    Returns:
      A list of 3-tuples of the form:
        (solver name, marginal contribution, best score w/o solver).
    """
    contribution_data = performance_data.marginal_contribution(aggregation_function,
                                                               minimise)
    contribution_data.sort(key=lambda contribution: contribution[1], reverse=True)
    return contribution_data


def compute_actual_selector_performance(
        actual_portfolio_selector: Path,
        performance_data: PerformanceDataFrame,
        feature_data: FeatureDataFrame,
        minimise: bool,
        aggregation_function: Callable[[list[float]], float],
        performance_cutoff: int | float = None) -> float:
    """Return the performance of the selector over all instances.

    Args:
      actual_portfolio_selector: Path to portfolio selector.
      performance_data: The performance data.
      feature_data: The feature data.
      minimise: Flag indicating, if scores should be minimised.
      aggregation_function: function to aggregate the performance per instance
      performance_cutoff: Optional performance cutoff.

    Returns:
      The selector performance as a single floating point number.
    """
    penalty_factor = gv.settings().get_general_penalty_multiplier()
    performances = []
    perf_measure = gv.settings().get_general_sparkle_objectives()[0].PerformanceMeasure
    for instance in performance_data.instances:
        # We get the performance for an instance by infering the model predicition
        # for the instance.
        performance_instance, flag_success = compute_actual_performance_for_instance(
            actual_portfolio_selector, instance, feature_data,
            performance_data, minimise, perf_measure, performance_cutoff)

        if not flag_success and performance_cutoff is not None:
            performance_instance = performance_cutoff * penalty_factor

        performances.append(performance_instance)

    return aggregation_function(performances)


def compute_actual_performance_for_instance(
        actual_portfolio_selector: Path,
        instance: str,
        feature_data: FeatureDataFrame,
        performance_data: PerformanceDataFrame,
        minimise: bool,
        objective_type: PerformanceMeasure,
        performance_cutoff: float = None) -> tuple[float, bool]:
    """Return the actual performance of the selector on a given instance.

    Args:
      actual_portfolio_selector: Path to the portfolio selector.
      instance: Instance name.
      feature_data: The feature data.
      performance_data: The Performance data
      minimise: Whether the performance value should be minimized or maximized
      objective_type: Whether we are dealing with run time or not.
      performance_cutoff: Cutoff value for this instance

    Returns:
      A 2-tuple where the first entry is the performance measure and
      the second entry is a Boolean indicating whether the instance was solved
      within the cutoff time (Runtime) or at least one solver performance
      did not exceed capvalue
    """
    # Get the prediction of the selector over the solvers
    selector = gv.settings().get_general_sparkle_selector()
    feature_vector = feature_data.get_instance(instance)
    predict_schedule = selector.run(actual_portfolio_selector, feature_vector)
    compare = operator.lt if minimise else operator.gt

    performance = None
    flag_successfully_solving = False
    if objective_type == PerformanceMeasure.RUNTIME:
        performance_list = []
        # In case of Runtime, we loop through the selected solvers
        for solver, schedule_cutoff in predict_schedule:
            # A prediction is a solver and its score given by the Selector
            performance = float(performance_data.get_value(solver, instance))
            performance_list.append(performance)
            # 1. if performance <= predicted runtime we have a successfull solver
            if performance <= schedule_cutoff:
                # 2. Success if the tried solvers < selector cutoff_time
                if sum(performance_list) <= performance_cutoff:
                    flag_successfully_solving = True
                break
            # 3. Else, we set the failed solver to the cutoff time
            performance_list[-1] = schedule_cutoff
            performance = sum(performance_list)
            # 4. If we have exceeded cutoff_time, we are done
            if performance_cutoff is not None and performance > performance_cutoff:
                break
    else:
        # Minimum or maximum of predicted solvers
        for solver, _ in predict_schedule:
            solver_performance = float(performance_data.get_value(solver, instance))
            if performance is None or compare(solver_performance, performance):
                performance = solver_performance

    return performance, flag_successfully_solving


def compute_actual_selector_marginal_contribution(
        performance_data: PerformanceDataFrame,
        feature_data: FeatureDataFrame,
        aggregation_function: Callable[[list[float]], float] = mean,
        performance_cutoff: int | float = None,
        minimise: bool = True,
        flag_recompute: bool = False,
        selector_timeout: int = 172000) -> list[tuple[str, float]]:
    """Compute the marginal contributions of solvers in the selector.

    Args:
      performance_data: Performance data object
      aggregation_function: Function to aggregate score values
      minimise: Flag indicating if scores should be minimised
      performance_cutoff: The cutoff performance of a solver
      feature_data_csv_path: Path to the CSV file with the feature data.
      flag_recompute: Boolean indicating whether marginal contributions should
        be recalculated even if they already exist in a file. Defaults to False.

    Returns:
      A list of 2-tuples where every 2-tuple is of the form
        (solver name, marginal contribution, best_performance).
    """
    actual_margi_cont_path = gv.settings().DEFAULT_marginal_contribution_actual_path
    # If the marginal contribution already exists in file, read it and return
    if not flag_recompute and actual_margi_cont_path.is_file():
        print("Marginal contribution for the actual selector already computed, reading "
              "from file instead! Use --recompute to force recomputation.")
        return ast.literal_eval(actual_margi_cont_path.open().read())

    print(f"In this calculation, cutoff is {performance_cutoff} seconds")

    # Compute performance of actual selector
    # NOTE: Should we recompute for all solvers?
    print("Computing actual performance for portfolio selector with all solvers ...")
    actual_portfolio_selector_path = gv.settings().DEFAULT_algorithm_selector_path
    construct_sparkle_portfolio_selector(actual_portfolio_selector_path,
                                         performance_data,
                                         feature_data,
                                         selector_timeout=selector_timeout)

    if not actual_portfolio_selector_path.exists():
        print(f"****** ERROR: {actual_portfolio_selector_path} does not exist! ******")
        print("****** ERROR: AutoFolio constructing the actual portfolio selector with"
              " all solvers failed! ******")
        print("****** Use virtual best performance instead of actual "
              "performance for this portfolio selector! ******")
        sys.exit(-1)

    actual_selector_performance = compute_actual_selector_performance(
        actual_portfolio_selector_path, performance_data,
        feature_data, minimise, aggregation_function, performance_cutoff)

    print("Actual performance for portfolio selector with all solvers is "
          f"{actual_selector_performance}")

    rank_list = []
    compare = operator.lt if minimise else operator.gt
    # Compute contribution per solver
    # NOTE: This could be parallelised
    for solver in performance_data.solvers:
        solver_name = Path(solver).name
        print("Computing actual performance for portfolio selector excluding solver "
              f"{solver_name} ...")
        # 1. Copy the dataframe original df
        tmp_performance_df = performance_data.copy()
        # 2. Remove the solver from this copy
        tmp_performance_df.remove_solver(solver)
        # 3. create the actual selector path
        tmp_actual_portfolio_selector = (
            gv.settings().DEFAULT_selection_output / f"without_{solver_name}"
            / "sparkle_portfolio_selector")

        if tmp_actual_portfolio_selector.exists():
            tmp_actual_portfolio_selector.unlink()
        else:
            tmp_actual_portfolio_selector.parent.mkdir(parents=True, exist_ok=True)

        if tmp_performance_df.num_solvers >= 1:
            # 4. Construct the portfolio selector for this subset
            construct_sparkle_portfolio_selector(
                tmp_actual_portfolio_selector, tmp_performance_df,
                feature_data)
        else:
            print("****** WARNING: No solver exists ! ******")

        if not tmp_actual_portfolio_selector.exists():
            sys.exit(-1)

        tmp_asp = compute_actual_selector_performance(
            tmp_actual_portfolio_selector, tmp_performance_df,
            feature_data, minimise, aggregation_function, performance_cutoff)

        print(f"Actual performance for portfolio selector ex. solver {solver_name} is "
              f"{tmp_asp}")
        print("Computing done!")

        # 1. If the performance remains equal, this solver did not contribute
        # 2. If there is a performance decay without this solver, it does contribute
        # 3. If there is a performance improvement, we have a bad portfolio selector
        if tmp_asp == actual_selector_performance:
            marginal_contribution = 0.0
        elif not compare(tmp_asp, actual_selector_performance):
            # In the case that the performance decreases, we have a contributing solver
            marginal_contribution = tmp_asp / actual_selector_performance
        else:
            print("****** WARNING DUBIOUS SELECTOR/SOLVER: "
                  f"The omission of solver {solver_name} yields an improvement. "
                  "The selector improves better without this solver. It may be usefull "
                  "to construct a portfolio without this solver.")
            marginal_contribution = 0.0

        rank_list.append((solver, marginal_contribution, tmp_asp))
        print(f"Marginal contribution (to Actual Selector) for solver {solver_name} is "
              f"{marginal_contribution}")

    rank_list.sort(key=lambda contribution: contribution[1], reverse=True)

    # Write actual selector contributions to file
    actual_margi_cont_path.open("w").write(str(rank_list))
    return rank_list


def compute_marginal_contribution(
        flag_compute_perfect: bool, flag_compute_actual: bool,
        flag_recompute: bool, selector_timeout: int) -> None:
    """Compute the marginal contribution.

    Args:
        flag_compute_perfect: Flag indicating if the contribution for the perfect
            portfolio selector should be computed.
        flag_compute_actual: Flag indicating if the contribution for the actual portfolio
             selector should be computed.
        flag_recompute: Flag indicating whether marginal contributions
            should be recalculated.
        selector_timeout: The cuttoff time to configure the algorithm selector.
    """
    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)
    performance_measure =\
        gv.settings().get_general_sparkle_objectives()[0].PerformanceMeasure
    aggregation_function = gv.settings().get_general_metric_aggregation_function()
    capvalue = gv.settings().get_general_cap_value()
    if performance_measure == PerformanceMeasure.QUALITY_ABSOLUTE_MAXIMISATION:
        minimise = False
    elif performance_measure == PerformanceMeasure.QUALITY_ABSOLUTE_MINIMISATION:
        minimise = True
    else:
        # assume runtime optimization
        capvalue = gv.settings().get_general_target_cutoff_time()
        minimise = True

    if not (flag_compute_perfect | flag_compute_actual):
        print("ERROR: compute_marginal_contribution called without a flag set to"
              " True, stopping execution")
        sys.exit(-1)

    if flag_compute_perfect:
        print("Computing each solver's marginal contribution to perfect selector ...")
        contribution_data = compute_perfect_selector_marginal_contribution(
            performance_data,
            aggregation_function,
            minimise)
        table = tabulate.tabulate(
            contribution_data,
            headers=["Solver", "Marginal Contribution", "Best Performance"],)
        print(table, "\n")
        print("Marginal contribution (perfect selector) computing done!")

    if flag_compute_actual:
        print("Start computing marginal contribution per Solver to actual selector...")
        contribution_data = compute_actual_selector_marginal_contribution(
            performance_data,
            feature_data,
            aggregation_function,
            capvalue,
            minimise,
            flag_recompute=flag_recompute,
            selector_timeout=selector_timeout
        )
        table = tabulate.tabulate(
            contribution_data,
            headers=["Solver", "Marginal Contribution", "Best Performance"],)
        print(table, "\n")
        print("Marginal contribution (actual selector) computing done!")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for SMAC2."""
import global_variables as gv
from sparkle.types.objective import PerformanceMeasure


def get_smac_run_obj() -> str:
    """Return the SMAC run objective.

    Returns:
        A string that represents the run objective set in the settings.
    """
    # Get smac_run_obj from general settings
    smac_run_obj = gv.settings().get_general_sparkle_objectives()[0].PerformanceMeasure

    # Convert to SMAC format
    if smac_run_obj == PerformanceMeasure.RUNTIME:
        return smac_run_obj.name
    elif smac_run_obj == PerformanceMeasure.QUALITY_ABSOLUTE_MINIMISATION:
        return "QUALITY"
    elif smac_run_obj == PerformanceMeasure.QUALITY_ABSOLUTE_MAXIMISATION:
        print("Warning: Performance measure not available for SMAC: {smac_run_obj}")
    else:
        print(f"Warning: Unknown SMAC objective {smac_run_obj}")
    return smac_run_obj



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions to run solvers."""
from __future__ import annotations
from pathlib import Path

from sparkle.CLI.help import global_variables as gv
from sparkle.platform.settings_objects import SolutionVerifier
from sparkle.solver import Solver
from sparkle.solver import sat_help as sh
import sparkle.tools.general as tg
from sparkle.tools.runsolver_parsing import handle_timeouts
from sparkle.types import SolverStatus


def run_solver_on_instance_and_process_results(
        solver: Solver, instance: Path | list[Path], custom_cutoff: int,
        seed: int) -> tuple[float, float, float, list[float], str, Path]:
    """Prepare and run a given the solver and instance, and process output.

    Args:
        solver: The solver to run on the instance
        instance: The path(s) to the instance file(s)
        custom_cutoff: The cutoff time for the solver
        seed: The seed for the solver

    Returns:
        tuple of the form:
            (cpu_time, wc_time, runtime, cpu_times, status, raw_result_path)
    """
    # Prepare paths
    if isinstance(instance, list):
        instance_name = instance[0].name
    else:
        instance_name = instance.name

    # Prepare runsolver call
    raw_result_path = solver.raw_output_directory /\
        f"{solver.name}_{instance_name}_{tg.get_time_pid_random_string()}.rawres"
    runsolver_watch_data_path = raw_result_path.with_suffix(".log")
    runsolver_values_path = raw_result_path.with_suffix(".val")
    solver_output = solver.run(
        instance,
        configuration={"seed": seed,
                       "cutoff_time": custom_cutoff,
                       "specifics": ""},
        runsolver_configuration=["--timestamp", "--use-pty",
                                 "--cpu-limit", str(custom_cutoff),
                                 "-w", runsolver_watch_data_path,
                                 "-v", runsolver_values_path,
                                 "-o", raw_result_path],
        cwd=Path.cwd())

    cpu_time_penalised, status =\
        handle_timeouts(solver_output["runtime"],
                        solver_output["status"],
                        custom_cutoff,
                        gv.settings().get_penalised_time(custom_cutoff))
    status = verify(instance, raw_result_path, solver, status)
    return (solver_output["cpu_time"], solver_output["wc_time"],
            cpu_time_penalised, solver_output["quality"], status, raw_result_path)


def verify(instance: Path, raw_result: Path, solver: Solver, status: str)\
        -> str:
    """Run a solution verifier on the solution and update the status if needed."""
    verifier = gv.settings().get_general_solution_verifier()
    # Use verifier if one is given and the solver did not time out
    if verifier == SolutionVerifier.SAT and status != SolverStatus.TIMEOUT \
            and status != SolverStatus.UNKNOWN:
        return sh.sat_verify(instance, raw_result, solver)
    return status



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import sys
import ast
import time
import random
import subprocess
from pathlib import Path

from sparkle.solver import Solver
from sparkle.tools.runsolver_parsing import get_runtime


if __name__ == "__main__":
    # Incoming call from SMAC:
    # 1. Translate input from SMAC to standardized form
    argsiter = iter(sys.argv[7:])
    args = zip(argsiter, argsiter)
    args = {arg.strip("-"): val for arg, val in args}
    # Args 1-6 conditions of the run, the rest are configurations for the solver
    # First argument is the solver directory
    solver_dir = Path(sys.argv[1])
    args["solver_dir"] = str(solver_dir)
    args["instance"] = sys.argv[2]
    args["specifics"] = sys.argv[3]
    args["cutoff_time"] = float(sys.argv[4])
    args["run_length"] = int(sys.argv[5])
    args["seed"] = int(sys.argv[6])
    cutoff_time = int(args["cutoff_time"])
    # 2. Build Run Solver call
    runsolver_binary = solver_dir / "runsolver"
    log_timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
    log_randstamp = int(random.getrandbits(32))
    runsolver_watch_data_path = Path(f"runsolver_{log_randstamp}_{log_timestamp}.log")
    runsolver_var_data_path = Path(f"runsolver_{log_randstamp}_{log_timestamp}.var")

    runsolver_call = [str(runsolver_binary),
                      "-w", str(runsolver_watch_data_path),
                      "-v", str(runsolver_var_data_path),
                      "--cpu-limit", str(cutoff_time),
                      str(solver_dir / Solver.wrapper),
                      '"' + str(args) + '"']
    # 3. Call Runsolver with the solver configurator wrapper and its arguments
    start_t = time.time()
    run_solver = subprocess.run(runsolver_call,
                                capture_output=True)
    run_time = min(time.time() - start_t, cutoff_time)

    # 4. Decode solver output and return required values to SMAC.
    # Solver output can be found in the regular subprocess.stdout
    if run_solver.returncode != 0:
        # Failure from run solver or solver wrapper
        print("WARNING: Subprocess for Solver Wrapper crashed with code "
              f"{run_solver.returncode}:\n {run_solver.stderr}")
        print(f"Result for SMAC: CRASHED, {run_time}, 0, 0, {args['seed']}")
        sys.exit()

    try:
        if run_solver.stdout.decode() == "":
            # Runsolver cutoff solver wrapper, default values
            outdict = {"status": "TIMEOUT"}
        else:
            outdict = ast.literal_eval(run_solver.stdout.decode())
    except Exception as ex:
        print("ERROR: Could not decode Solver Wrapper output:\n"
              f"Return code: {run_solver.returncode}"
              f"stdout: {run_solver.stdout}\n "
              f"stderr: {run_solver.stderr}\n "
              f"Exception whilst decoding dictionary: {ex}")
        sys.exit(-1)

    # Overwrite the CPU runtime with runsolver log value
    # TODO: Runsolver also registers WALL time, add as a settings option in Sparkle
    runsolver_runtime, run_wtime = get_runtime(runsolver_var_data_path)
    if runsolver_runtime != -1.0:  # Valid value found
        run_time = runsolver_runtime
        runsolver_var_data_path.unlink(missing_ok=True)
    elif run_wtime != -1.0:
        run_time = run_wtime
        print("WARNING: CPU time not found in Runsolver log. "
              "Using Runsolver Wall Time instead.")
    else:
        print("WARNING: Was not able to deduce runtime from Runsolver. Using Python "
              f"timer instead for runtime. See {runsolver_var_data_path}")

    # 5. Return values to SMAC
    # We need to check how the "quality" in the output directory must be formatted
    quality = "0"
    if "quality" in outdict.keys():
        quality = outdict["quality"]
        if isinstance(quality, dict):
            # SMAC2 does not support multi-objective so always opt for the first value
            quality = quality[quality.keys()[0]]

    print("Result for SMAC: "
          f"{outdict['status']}, {run_time}, 0, {quality}, {args['seed']}")



#!/usr/bin/env python3
"""Sparkle command to construct a portfolio selector."""
import sys
import argparse
from pathlib import Path

from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.CLI.support import compute_marginal_contribution_help as scmch
from sparkle.CLI.help import sparkle_logging as sl
from sparkle.platform.settings_objects import SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.types.objective import PerformanceMeasure


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.RecomputePortfolioSelectorArgument.names,
                        **ac.RecomputePortfolioSelectorArgument.kwargs)
    parser.add_argument(*ac.RecomputeMarginalContributionForSelectorArgument.names,
                        **ac.RecomputeMarginalContributionForSelectorArgument.kwargs)
    parser.add_argument(*ac.SelectorTimeoutArgument.names,
                        **ac.SelectorTimeoutArgument.kwargs)
    parser.add_argument(*ac.PerformanceMeasureArgument.names,
                        **ac.PerformanceMeasureArgument.kwargs)

    return parser


def judge_exist_remaining_jobs(feature_data_csv: Path,
                               performance_data_csv: Path) -> bool:
    """Return whether there are remaining feature or performance computation jobs."""
    feature_data = FeatureDataFrame(feature_data_csv)
    if feature_data.has_missing_vectors():
        return True
    performance_data = PerformanceDataFrame(performance_data_csv)
    return performance_data.has_missing_performance()


def construct_sparkle_portfolio_selector(selector_path: Path,
                                         performance_data: PerformanceDataFrame,
                                         feature_data: FeatureDataFrame,
                                         selector_timeout: int = None) -> None:
    """Create the Sparkle portfolio selector.

    Applies several checks before executing, such as missing data imputation.

    Args:
        selector_path: Portfolio selector path.
        performance_data: Performance data.
        feature_data: Feature data.
        flag_recompute: Whether or not to recompute if the selector exists and no data
            was changed. Defaults to False.
        selector_timeout: The cuttoff time to configure the algorithm selector. If None
            uses the default selector configuration. Defaults to None.
    """
    cutoff_time = gv.settings().get_general_target_cutoff_time()

    # Selector (AutoFolio) cannot handle cutoff time less than 2, adjust if needed
    cutoff_time = max(cutoff_time, 2)

    # Determine the objective function
    perf_measure = gv.settings().get_general_sparkle_objectives()[0].PerformanceMeasure
    if perf_measure == PerformanceMeasure.RUNTIME:
        objective_function = "runtime"
    elif perf_measure == PerformanceMeasure.QUALITY_ABSOLUTE_MAXIMISATION or\
            perf_measure == PerformanceMeasure.QUALITY_ABSOLUTE_MINIMISATION:
        objective_function = "solution_quality"
    else:
        print("ERROR: Unknown performance measure in portfolio selector construction.")
        sys.exit(-1)

    if feature_data.has_missing_value():
        print("****** WARNING: There are missing values in the feature data, and all "
              "missing values will be imputed as the mean value of all other non-missing"
              " values! ******")
        print("Imputing all missing values...")
        feature_data.impute_missing_values()

    selector = gv.settings().get_general_sparkle_selector()
    selector_path = selector.construct(selector_path,
                                       performance_data,
                                       feature_data,
                                       objective_function,
                                       cutoff_time,
                                       selector_timeout)


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    flag_recompute_portfolio = args.recompute_portfolio_selector
    flag_recompute_marg_cont = args.recompute_marginal_contribution

    check_for_initialise(
        COMMAND_DEPENDENCIES[CommandName.CONSTRUCT_SPARKLE_PORTFOLIO_SELECTOR]
    )

    if ac.set_by_user(args, "performance_measure"):
        gv.settings().set_general_sparkle_objectives(
            args.performance_measure, SettingState.CMD_LINE
        )

    print("Start constructing Sparkle portfolio selector ...")

    flag_judge_exist_remaining_jobs = judge_exist_remaining_jobs(
        gv.settings().DEFAULT_feature_data_path,
        gv.settings().DEFAULT_performance_data_path)

    if flag_judge_exist_remaining_jobs:
        print("There remain unperformed feature computation jobs or performance "
              "computation jobs!")
        print("Please first execute all unperformed jobs before constructing Sparkle "
              "portfolio selector")
        print("Sparkle portfolio selector is not successfully constructed!")
        sys.exit(-1)

    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)

    selector_path = gv.settings().DEFAULT_algorithm_selector_path
    if selector_path.exists() and not flag_recompute_portfolio:
        print("Portfolio selector already exists. Set the recompute flag to re-create.")
    else:
        construct_sparkle_portfolio_selector(
            selector_path,
            performance_data,
            feature_data,
            args.selector_timeout
        )
        print("Sparkle portfolio selector constructed!")

    print(f"Sparkle portfolio selector located at {selector_path}")

    # Update latest scenario
    gv.latest_scenario().set_selection_portfolio_path(selector_path)
    gv.latest_scenario().set_latest_scenario(Scenario.SELECTION)
    # Set to default to overwrite possible old path
    gv.latest_scenario().set_selection_test_case_directory()

    # Compute and print marginal contributions of the perfect and actual portfolio
    # selectors
    scmch.compute_marginal_contribution(flag_compute_perfect=True,
                                        flag_compute_actual=True,
                                        flag_recompute=flag_recompute_marg_cont,
                                        selector_timeout=args.selector_timeout)

    # Write used settings to file
    gv.settings().write_used_settings()
    # Write used scenario to file
    gv.latest_scenario().write_scenario_ini()



"""Methods related to SAT specific runs."""
from pathlib import Path
import subprocess

from sparkle.solver import Solver

sat_verifier_path = Path("sparkle/Components/Sparkle-SAT-verifier/SAT")


def sat_verify(instance: Path, raw_result: Path, solver: Solver) -> str:
    """Run a SAT verifier and return its status."""
    status = sat_judge_correctness_raw_result(instance, raw_result)

    if status != "SAT" and status != "UNSAT" and status != "WRONG":
        status = "UNKNOWN"
        print(f"Warning: Verification result was UNKNOWN for solver {solver.name} on "
              f"instance {instance.name}!")

    # TODO: Make removal conditional on a success status (SAT or UNSAT)
    # sfh.rmfiles(raw_result_path)
    return status


def sat_get_verify_string(sat_output: str) -> str:
    """Return the status of the SAT verifier.

    Four statuses are possible: "SAT", "UNSAT", "WRONG", "UNKNOWN"
    """
    lines = [line.strip() for line in sat_output.splitlines()]
    for index, line in enumerate(lines):
        if line == "Solution verified.":
            if lines[index + 2] == "11":
                return "SAT"
        elif line == "Solver reported unsatisfiable. I guess it must be right!":
            if lines[index + 2] == "10":
                return "UNSAT"
        elif line == "Wrong solution.":
            if lines[index + 2] == "0":
                return "WRONG"
    return "UNKNOWN"


def sat_judge_correctness_raw_result(instance: Path, raw_result: Path) -> str:
    """Run a SAT verifier to determine correctness of a result.

    Args:
        instance: path to the instance
        raw_result: path to the result to verify

    Returns:
        The status of the solver on the instance
    """
    print("Run SAT verifier")
    sat_verify = subprocess.run([sat_verifier_path, instance, raw_result],
                                capture_output=True)
    print("SAT verifier done")
    return sat_get_verify_string(sat_verify.stdout.decode())



"""Class for solver status."""
from enum import Enum


class SolverStatus(str, Enum):
    """Possible return states for solver runs."""
    SUCCESS = "SUCCESS"
    CRASHED = "CRASHED"
    TIMEOUT = "TIMEOUT"
    UNKNOWN = "UNKNOWN"
    ERROR = "ERROR"
    KILLED = "KILLED"


def get_status_from_str(status: str) -> SolverStatus:
    """Get SolverStatus object from string."""
    return SolverStatus(status.upper())



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for parallel portfolio report generation."""
from pathlib import Path

from sparkle.platform import generate_report_for_selection as sgfs
from sparkle.types.objective import PerformanceMeasure, SparkleObjective
from sparkle.instance import InstanceSet


def get_solver_list_latex(solver_list: list[str]) -> str:
    """Return the list of solvers as string, including each solver-seed combination.

    Args:
        solver_list: The solver list to convert

    Returns:
        A list of solvers in the parallel portfolio as str.
    """
    latex_itemize = ""

    for solver_str in solver_list:
        # Solver string may contain path and variation seed
        solver_split = solver_str.split(" ")
        solver_name = Path(solver_split[0]).name
        solver_seeds = int(solver_split[2]) if len(solver_split) == 3 else 0

        latex_itemize += f"\\item \\textbf{{{sgfs.underscore_for_latex(solver_name)}}}\n"
        # Only include if we used more than one seed
        if solver_seeds > 1:
            seeds = ",".join(list[range(1, solver_seeds + 1)])
            latex_itemize += f"\\item[]With seeds: {seeds}\n"

    return latex_itemize


def get_dict_sbs_penalty_time_on_each_instance(
        solver_list: list[str],
        instance_set: InstanceSet,
        results: dict[list[str, str]],
        cutoff: int,
        penalised_time: int) -> tuple[dict[str, float], str, dict[str, float]]:
    """Return the penalised run time for the single best solver and per solver.

    Args:
        solver_list: list of solvers
        instance_list: List of paths to instance sets
        results: dictionary of results of the portfolio

    Returns:
        A three-tuple:
            A dict containing the run time per instance for the single best solver.
            A string containing the name of the single best solver.
            A second dict containing penalised averaged run time per solver.
    """
    penalised_averaged_runtime = {solver: 0.0 for solver in solver_list}
    penalised_time = float(penalised_time)
    for instance in results:
        for (solver, status, runtime) in results[instance]:
            # Those that did not finish or exceed the thresh get the penalty
            if float(runtime) >= cutoff or status.lower() != "success":
                runtime = penalised_time
            penalised_averaged_runtime[solver] += (float(runtime) / instance_set.size)

    # Find the single best solver (SBS)
    sbs_name = min(penalised_averaged_runtime, key=penalised_averaged_runtime.get)
    sbs_name = Path(sbs_name).name
    sbs_dict = {name: penalised_time for name in instance_set._instance_names}

    for instance in results:
        instance_name = Path(instance).name
        found = False
        for result in results[instance_name]:
            if result[0] == sbs_name:
                found = True
                if float(result[2]) < penalised_time:
                    sbs_dict[instance_name] = float(result[2])
                break
        if not found:
            print(f"WARNING: No result found for instance: {instance_name}")
    return sbs_dict, sbs_name, penalised_averaged_runtime


def get_dict_actual_parallel_portfolio_penalty_time_on_each_instance(
        instance_set: InstanceSet,
        results: dict[str, list[str]],
        cutoff_time: int,
        penalised_time: int) -> dict[str, float]:
    """Returns the instance names and corresponding penalised running times of the PaP.

    Args:
        instance_list: List of instances.

    Returns:
        A dict of instance names and the penalised running time of the PaP.
    """
    instance_penalty_dict = {}

    default_penalty = float(penalised_time)

    for instance_name in instance_set._instance_names:
        if instance_name in results:
            selected_value = results[instance_name][0]
            for solver in results[instance_name]:
                if solver[2] < selected_value[2]:
                    selected_value = solver
            _, status, runtime = selected_value
            # Assign the measured result if exit status was not timeout
            # and measured time less then cutoff
            if float(runtime) < cutoff_time and status != "TIMEOUT":
                instance_penalty_dict[instance_name] = float(runtime)
            else:  # Assign the penalized result
                instance_penalty_dict[instance_name] = default_penalty
        else:
            instance_penalty_dict[instance_name] = default_penalty
    return instance_penalty_dict


def get_figure_parallel_portfolio_sparkle_vs_sbs(
        target_directory: Path,
        solver_list: list[str],
        instance_set: InstanceSet,
        results: list[str],
        objective: SparkleObjective,
        cutoff: int,
        penalised_time: int) -> tuple[
        str, dict[str, float], dict[str, float]]:
    """Generate PaP vs SBS figure and return a string to include it in LaTeX.

    Args:
        target_directory: Path where to place the files
        parallel_portfolio_path: Parallel portfolio path.
        instances: List of instances.

    Returns:
        a three tuple:
            str_value: A string to include the PaP vs SBS figure in LaTeX
            dict_all_solvers: A dict containing the penalised average run time per
                solver.
            dict_actual_parallel_portfolio_penalty_time_on_each_instance: A dict with
                instance names and the penalised running time of the PaP.
    """
    dict_sbs_penalty_time_on_each_instance, sbs_solver, dict_all_solvers =\
        get_dict_sbs_penalty_time_on_each_instance(solver_list, instance_set, results,
                                                   cutoff, penalised_time)
    dict_actual_parallel_portfolio_penalty_time_on_each_instance =\
        get_dict_actual_parallel_portfolio_penalty_time_on_each_instance(instance_set,
                                                                         results,
                                                                         cutoff,
                                                                         penalised_time)

    figure_filename = "figure_parallel_portfolio_sparkle_vs_sbs"
    data_filename = "data_parallel_portfolio_sparkle_vs_sbs.dat"
    data_filepath = target_directory / data_filename

    with data_filepath.open("w") as outfile:
        for instance in dict_sbs_penalty_time_on_each_instance:
            sbs_penalty_time = dict_sbs_penalty_time_on_each_instance[instance]
            sparkle_penalty_time = (
                dict_actual_parallel_portfolio_penalty_time_on_each_instance[instance])
            outfile.write(str(sbs_penalty_time) + " " + str(sparkle_penalty_time) + "\n")

    generate_figure(target_directory, data_filename, float(penalised_time),
                    f"SBS ({sgfs.underscore_for_latex(sbs_solver)})",
                    "Parallel-Portfolio", figure_filename, objective.metric)
    latex_include = f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"
    return (latex_include, dict_all_solvers,
            dict_actual_parallel_portfolio_penalty_time_on_each_instance)


def get_results_table(results: dict[str, str, str],
                      dict_all_solvers: dict[str, float], parallel_portfolio_path: Path,
                      dict_portfolio: dict[str, float],
                      solver_with_solutions: dict[str, int],
                      n_unsolved_instances: int, n_instances: int,
                      performance_metric: str) -> str:
    """Returns a LaTeX table with the portfolio results.

    Args:
        results: The total results with status and runtime per solver
        dict_all_solvers: A dict containing the penalised average run time per solver.
        parallel_portfolio_path: Parallel portfolio path
        dict_portfolio: A dict with instance names and the penalised running time of the
            PaP.
        solver_with_solutions: A dict with solver name as key, and number of solved
            instances for the corresponding solver as value, see
            get_solvers_with_solution.
        n_unsolved_instances: Number of unsolved instances.
        n_instances: Number of instances.

    Returns:
        A string containing LaTeX code for a table with the portfolio results.
    """
    portfolio_par = 0.0
    for instance in dict_portfolio:
        portfolio_par += dict_portfolio[instance]
    portfolio_par = portfolio_par / n_instances
    total_killed = 0
    for instance in results:
        for (_, status, _) in results[instance]:
            total_killed += (status.lower() == "killed")
    # Table 1: Portfolio results
    table_string = (
        "\\caption{\\textbf{Portfolio results}} \\label{tab:portfolio_results} ")
    table_string += "\\begin{tabular}{rrrrr}"
    table_string += (
        "\\textbf{Portfolio nickname} & \\textbf{"
        f"{performance_metric}"
        "} & \\textbf{\\#Timeouts} & "
        "\\textbf{\\#Cancelled} & \\textbf{\\# Solved} \\\\ \\hline ")
    table_string += (
        f"{sgfs.underscore_for_latex(parallel_portfolio_path.name)} & "
        f"{round(portfolio_par,2)} & {n_unsolved_instances} & {total_killed} & "
        f"{n_instances-n_unsolved_instances} \\\\ ")
    table_string += "\\end{tabular}"
    table_string += "\\bigskip"
    # Table 2: Solver results
    table_string += "\\caption{\\textbf{Solver results}} \\label{tab:solver_results} "
    table_string += "\\begin{tabular}{rrrrr}"

    for i, line in enumerate(dict_all_solvers):
        solver_name = Path(line).name

        if i == 0:
            table_string += (
                "\\textbf{Solver} & \\textbf{"
                f"{performance_metric}"
                "} & \\textbf{\\#Timeouts} & "
                "\\textbf{\\#Cancelled} & \\textbf{\\#Best solver} \\\\ \\hline ")

        if solver_name not in solver_with_solutions:
            cancelled = n_instances - n_unsolved_instances
            table_string += (
                f"{sgfs.underscore_for_latex(solver_name)} & "
                f"{round(dict_all_solvers[line], 2)} & {n_unsolved_instances} & "
                f"{cancelled} & 0 \\\\ ")
        else:
            cancelled = (n_instances - n_unsolved_instances
                         - solver_with_solutions[solver_name])
            table_string += (
                f"{sgfs.underscore_for_latex(solver_name)} & "
                f"{round(dict_all_solvers[line], 2)} & {n_unsolved_instances} & "
                f"{cancelled} & {solver_with_solutions[solver_name]} \\\\ ")
    table_string += "\\end{tabular}"

    return table_string


def parallel_report_variables(target_directory: Path,
                              parallel_portfolio_path: Path,
                              bibliograpghy_path: Path,
                              objective: SparkleObjective,
                              cutoff: int,
                              penalised_time: int,
                              instance_set: InstanceSet) -> dict[str, str]:
    """Returns a mapping between LaTeX report variables and their values.

    Args:
        target_directory: Path to where to place the generated files.
        parallel_portfolio_path: Parallel portfolio path.
        bibliograpghy_path: Path to the bib file
        instances: List of instances.

    Returns:
        A dictionary that maps variables used in the LaTeX report to values.
    """
    variables_dict = {"bibliographypath":
                      str(bibliograpghy_path.absolute())}
    # Get the results data
    csv_data = [line.split(",") for line in
                (parallel_portfolio_path / "results.csv").open("r").readlines()]
    solver_list = list(set([line[1] for line in csv_data]))  # Unique set of solvers
    results = {name: [] for name in instance_set._instance_names}
    for row in csv_data:
        if row[0] in results.keys():
            results[row[0]].append([row[1], row[2], row[3]])

    variables_dict["numSolvers"] = str(len(solver_list))
    variables_dict["solverList"] = get_solver_list_latex(solver_list)
    variables_dict["numInstanceClasses"] = "1"  # We currently do not support multi sets
    variables_dict["cutoffTime"] = str(cutoff)
    variables_dict["performanceMetric"] = objective.metric
    variables_dict["instanceClassList"] =\
        sgfs.get_instance_set_count_list(instance_set.instance_paths)

    # Produce some statistics on the parallel portfolio
    solvers_solutions = {solver: 0 for solver in solver_list}
    instance_names_copy = instance_set._instance_names.copy()
    for line in csv_data:
        if line[0] in instance_names_copy and line[2].lower() == "success":
            solvers_solutions[line[1]] += 1
            instance_names_copy.remove(line[0])
    unsolved_instances = instance_set.size - sum([solvers_solutions[key]
                                                  for key in solvers_solutions])
    inst_succes = []
    for solver in solvers_solutions:
        inst_succes.append("\\item Solver "
                           f"\\textbf{{{sgfs.underscore_for_latex(solver)}}}, was the "
                           "best solver on "
                           f"\\textbf{{{solvers_solutions[solver]}}} instance(s)")
    if unsolved_instances > 0:
        inst_succes.append(f"\\item \\textbf{{{unsolved_instances}}} instances(s) "
                           "remained unsolved")

    variables_dict["solversWithSolution"] = "\n".join(inst_succes)

    (figure_name, dict_all_solvers,
        dict_actual_parallel_portfolio_penalty_time_on_each_instance) =\
        get_figure_parallel_portfolio_sparkle_vs_sbs(target_directory, solver_list,
                                                     instance_set, results, objective,
                                                     cutoff, penalised_time)

    variables_dict["figure-parallel-portfolio-sparkle-vs-sbs"] = figure_name

    variables_dict["resultsTable"] = get_results_table(
        results, dict_all_solvers, parallel_portfolio_path,
        dict_actual_parallel_portfolio_penalty_time_on_each_instance,
        solvers_solutions, unsolved_instances, instance_set.size, objective.metric)

    if objective.PerformanceMeasure == PerformanceMeasure.QUALITY_ABSOLUTE_MINIMISATION:
        variables_dict["decisionBool"] = "\\decisionfalse"
    else:
        variables_dict["decisionBool"] = "\\decisiontrue"
    return variables_dict


def generate_figure(
        target_directory: Path, data_parallel_portfolio_sparkle_vs_sbs_filename: str,
        penalty_time: float, sbs_name: str, parallel_portfolio_sparkle_name: str,
        figure_parallel_portfolio_sparkle_vs_sbs_filename: str,
        performance_measure: str) -> None:
    """Generates image for parallel portfolio report."""
    upper_bound = penalty_time * 1.5
    lower_bound = 0.01

    output_eps_file =\
        f"{figure_parallel_portfolio_sparkle_vs_sbs_filename}.eps"
    sbs_name = sbs_name.replace("/", "_")
    output_gnuplot_script = f"{parallel_portfolio_sparkle_name}_vs_{sbs_name}.plt"

    with (target_directory / output_gnuplot_script).open("w+") as outfile:
        outfile.write(f"set xlabel '{sbs_name}, {performance_measure}'\n"
                      f"set ylabel '{parallel_portfolio_sparkle_name}, "
                      f"{performance_measure}'\n"
                      f"set title '{parallel_portfolio_sparkle_name} vs {sbs_name}'\n"
                      "unset key\n"
                      f"set xrange [{lower_bound}:{upper_bound}]\n"
                      f"set yrange [{lower_bound}:{upper_bound}]\n"
                      "set logscale x\n"
                      "set logscale y\n"
                      "set grid\n"
                      "set size square\n"
                      f"set arrow from {lower_bound},{lower_bound} to {upper_bound},"
                      f"{upper_bound} nohead lc rgb 'black'\n")

        if performance_measure == "PAR10":
            # Cutoff time x axis
            outfile.write(f"set arrow from {penalty_time},{lower_bound} to "
                          f"{penalty_time},{upper_bound} nohead lc rgb 'black' lt 2\n")
            # Cutoff time y axis
            outfile.write(f"set arrow from {lower_bound},{penalty_time} to {upper_bound}"
                          f",{penalty_time} nohead lc rgb 'black' lt 2\n")

        outfile.write('set terminal postscript eps color dashed linewidth "Helvetica"'
                      " 20\n")
        outfile.write(f"set output '{output_eps_file}'\n"
                      f"plot '{data_parallel_portfolio_sparkle_vs_sbs_filename}' with "
                      "points pt 2 ps 2\n")

    sgfs.generate_gnuplot(output_gnuplot_script, target_directory)
    sgfs.generate_pdf(output_eps_file, target_directory)
    Path(output_gnuplot_script).unlink(missing_ok=True)


def generate_report_parallel_portfolio(parallel_portfolio_path: Path,
                                       target_path: Path,
                                       latex_template: Path,
                                       bibliograpghy_path: Path,
                                       objective: SparkleObjective,
                                       cutoff: int,
                                       penalised_time: int,
                                       instances: InstanceSet) -> None:
    """Generate a report for a parallel algorithm portfolio.

    Args:
        parallel_portfolio_path: Parallel portfolio path.
        target_path: Where the report data will be placed.
        latex_template: Path to the latex template path used
        bibliograpghy_path: Path to the bib file
        objective: The objective of the portfolio
        cutoff: The cutoff time for each solver
        penalised_time: The penalty for TIMEOUT solvers
        instances: List of instances.
    """
    target_path.mkdir(parents=True, exist_ok=True)
    dict_variable_to_value = parallel_report_variables(
        target_path, parallel_portfolio_path, bibliograpghy_path, objective,
        cutoff, penalised_time, instances)

    sgfs.generate_report(latex_template,
                         "template-Sparkle-for-parallel-portfolio.tex",
                         target_path,
                         "Sparkle_Report_Parallel_Portfolio",
                         dict_variable_to_value)



"""Setup file for Sparkle."""
import os
from setuptools import setup, find_packages
from sparkle.about import version

setup(name="SparkleAI",
      version=version,
      url="https://github.com/thijssnelleman/Sparkle",
      author="Thijs Snelleman",
      author_email="fkt_sparkle@aim.rwth-aachen.de",
      keywords="ai sat planning",
      description="Sparkle is a Programming by Optimisation (PbO)-based problem-solving "
                  "platform designed to enable the widespread and effective use of PbO "
                  "techniques for improving the state-of-the-art in solving a broad "
                  "range of prominent AI problems, including SAT and AI Planning.",
      long_description=open("README.MD", "r").read() if os.path.exists("README.MD") else "",
      long_description_content_type="text/markdown",
      packages=find_packages(exclude=["tests"]),
      include_package_data=True,
      entry_points={"console_scripts": ["sparkle=sparkle.CLI.cli:main"], },)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for selection report generation."""
import sys
import numpy as np
from shutil import which
from pathlib import Path
from collections import Counter
import subprocess

from sparkle.platform import tex_help as stex
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.CLI.support import compute_marginal_contribution_help as scmch
from sparkle.types.objective import PerformanceMeasure, SparkleObjective


def underscore_for_latex(string: str) -> str:
    """Return the input str with the underscores escaped for use in LaTeX.

    Args:
        string: A given str with underscores.

    Returns:
        The corresponding str with underscores escaped.
    """
    return string.replace("_", "\\_")


def get_solver_list_latex(solver_list: list[str] = None) -> str:
    """Get the list of solvers for use in a LaTeX document.

    Returns:
        The list of solver names as LaTeX str.
    """
    return "".join(f"\\item \\textbf{{{Path(solver_path).name}}}\n"
                   for solver_path in solver_list)


def get_feature_extractor_list(extractor_dir: Path) -> str:
    """Get the feature extractors for use in a LaTeX document.

    Returns:
        The list of feature extractors as LaTeX str.
    """
    return "".join(f"\\item \\textbf{{{Path(extractor_path).name}}}\n"
                   for extractor_path in extractor_dir.iterdir())


def get_num_instance_sets(instance_list: list[str]) -> str:
    """Get the number of instance sets.

    Args:
        instance_list: List of instances to use

    Returns:
        The number of instance sets as LaTeX str.
    """
    return str(len(set([Path(instance_path).parent.name
                        for instance_path in instance_list])))


def get_instance_set_count_list(instance_list: list[str] = None) -> str:
    """Get the instance sets for use in a LaTeX document.

    Returns:
        The list of instance sets as LaTeX str.
    """
    instance_list = [Path(instance_path).parent.name
                     for instance_path in instance_list]
    count = Counter(instance_list)
    return "".join(f"\\item \\textbf{ {inst_key} }, consisting of {count[inst_key]} "
                   "instances\n" for inst_key in count)


def solver_rank_list_latex(rank_list: list[tuple[str, float]]) -> str:
    """Convert solvers ranked by marginal contribution to latex.

    Returns:
        Solvers in the VBS (virtual best solver) ranked by marginal contribution as LaTeX
        str.
    """
    return "".join(f"\\item \\textbf{ {Path(solver).name} }, marginal contribution: "
                   f"{value}\n" for solver, value, _ in rank_list)


def get_par_ranking_list(performance_data: PerformanceDataFrame,
                         objective: SparkleObjective) -> str:
    """Get a list of the solvers ranked by PAR (Penalised Average Runtime).

    Returns:
        The list of solvers ranked by PAR as LaTeX str.
    """
    solver_penalty_ranking = performance_data.get_solver_penalty_time_ranking()
    return "".join(f"\\item \\textbf{{{solver}}}, {objective.metric}: {solver_penalty}\n"
                   for solver, solver_penalty in solver_penalty_ranking)


def get_actual_par(performance_dict: dict) -> str:
    """PAR (Penalised Average Runtime) of the Sparkle portfolio selector.

    Returns:
        The PAR (Penalised Average Runtime) of the Sparkle portfolio selector over a set
        of instances.
    """
    mean_performance = sum(performance_dict.values()) / len(performance_dict)
    return str(mean_performance)


def get_test_actual_par(performance_data: PerformanceDataFrame) -> str:
    """Return the true PAR (Penalised Average Runtime) score on a test set.

    Args:
        test_case_directory: Path to the test case directory.

    Returns:
        PAR score (Penalised Average Runtime) as string.
    """
    # Its selecting the first solver because in this case its the Selector (Only solver)
    solver = performance_data.solvers[0]
    return str(performance_data.dataframe[solver].sum() / performance_data.num_instances)


def get_dict_sbs_penalty_time_on_each_instance(
        performance_data: PerformanceDataFrame) -> dict[str, int]:
    """Returns a dictionary with the penalised performance of the Single Best Solver.

    Returns:
        A dict that maps instance name str to their penalised performance int.
    """
    solver_penalty_time_ranking_list =\
        performance_data.get_solver_penalty_time_ranking()
    sbs_solver = solver_penalty_time_ranking_list[0][0]
    return {instance: performance_data.get_value(sbs_solver, instance)
            for instance in performance_data.instances}


def get_actual_portfolio_selector_performance_per_instance(
        performance_data: PerformanceDataFrame,
        actual_portfolio_selector_path: Path,
        feature_data: FeatureDataFrame,
        capvalue: int,
        penalised_time: int) -> dict[str, int]:
    """Creates a dictionary with the portfolio selector performance on each instance.

    Returns:
        A dict that maps instance name str to their penalised performance int.
    """
    objective = SparkleObjective(performance_data.objective_names[0])
    minimise =\
        objective.PerformanceMeasure != PerformanceMeasure.QUALITY_ABSOLUTE_MAXIMISATION

    actual_selector_penalty = {}
    for instance in performance_data.instances:
        used_time_for_this_instance, flag_successfully_solving = \
            scmch.compute_actual_performance_for_instance(
                actual_portfolio_selector_path, instance, feature_data,
                performance_data, minimise, objective.PerformanceMeasure, capvalue)

        if flag_successfully_solving:
            actual_selector_penalty[instance] = used_time_for_this_instance
        else:
            actual_selector_penalty[instance] = penalised_time

    return actual_selector_penalty


def get_figure_portfolio_selector_sparkle_vs_sbs(output_dir: Path,
                                                 objective: SparkleObjective,
                                                 train_data: PerformanceDataFrame,
                                                 actual_portfolio_selector_penalty: dict,
                                                 penalty: int) -> str:
    """Create a LaTeX plot comparing the selector and the SBS.

    The plot compares the performance on each instance of the portfolio selector created
    by Sparkle and the SBS (single best solver).

    Returns:
        LaTeX str to include the comparison plot in a LaTeX report.
    """
    sbs_penalty_time = get_dict_sbs_penalty_time_on_each_instance(train_data)

    instances = sbs_penalty_time.keys() & actual_portfolio_selector_penalty.keys()
    if (len(sbs_penalty_time) != len(instances)):
        print("ERROR: Number of penalty times for the single best solver does not match "
              "the number of instances")
        sys.exit(-1)
    points = [[sbs_penalty_time[instance], actual_portfolio_selector_penalty[instance]]
              for instance in instances]

    figure_filename = "figure_portfolio_selector_sparkle_vs_sbs"

    solver_penalty_time_ranking_list =\
        train_data.get_solver_penalty_time_ranking()
    sbs_solver = Path(solver_penalty_time_ranking_list[0][0]).name

    generate_comparison_plot(points,
                             figure_filename,
                             xlabel=f"SBS ({sbs_solver}) [{objective.metric}]",
                             ylabel=f"Sparkle Selector [{objective.metric}]",
                             limit="magnitude",
                             limit_min=0.25,
                             limit_max=0.25,
                             penalty_time=penalty,
                             replace_zeros=True,
                             output_dir=output_dir)
    return f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"


def get_figure_portfolio_selector_sparkle_vs_vbs(output_dir: Path,
                                                 objective: SparkleObjective,
                                                 train_data: PerformanceDataFrame,
                                                 actual_portfolio_selector_penalty: dict,
                                                 penalty: int) -> str:
    """Create a LaTeX plot comparing the selector and the VBS.

    The plot compares the performance on each instance of the portfolio selector created
    by Sparkle and the VBS (virtual best solver).

    Returns:
        LaTeX str to include the comparison plot in a LaTeX report.
    """
    vbs_penalty_time = train_data.get_dict_vbs_penalty_time_on_each_instance(penalty)

    instances = vbs_penalty_time.keys() & actual_portfolio_selector_penalty.keys()
    if (len(vbs_penalty_time) != len(instances)):
        print("ERROR: Number of penalty times for the virtual best solver does not"
              "match the number of instances")
        sys.exit(-1)
    points = [[vbs_penalty_time[instance], actual_portfolio_selector_penalty[instance]]
              for instance in instances]

    figure_filename = "figure_portfolio_selector_sparkle_vs_vbs"

    generate_comparison_plot(points,
                             figure_filename,
                             xlabel=f"VBS [{objective.metric}]",
                             ylabel=f"Sparkle Selector [{objective.name}]",
                             limit="magnitude",
                             limit_min=0.25,
                             limit_max=0.25,
                             penalty_time=penalty,
                             replace_zeros=True,
                             output_dir=output_dir)

    return f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"


def selection_report_variables(
        target_dir: Path,
        bibliograpghy_path: Path,
        extractor_path: Path,
        actual_portfolio_selector_path: Path,
        feature_data: FeatureDataFrame,
        extractor_cutoff: int,
        cutoff: int,
        penalty: int,
        train_data: PerformanceDataFrame,
        test_case_data: PerformanceDataFrame = None) -> dict[str, str]:
    """Returns: a dict matching variables in the LaTeX template with their values.

    Args:
        target_dir: Output path
        bibliography_path: Path to the bib file
        test_case_directory: Path to the test case directory.

    Returns:
        A dict matching str variables in the LaTeX template with their value str.
    """
    objective = SparkleObjective(train_data.objective_names[0])
    actual_performance_dict = get_actual_portfolio_selector_performance_per_instance(
        train_data, actual_portfolio_selector_path, feature_data, cutoff, penalty)
    latex_dict = {"bibliographypath": str(bibliograpghy_path.absolute()),
                  "numSolvers": str(train_data.num_solvers),
                  "solverList": get_solver_list_latex(train_data.solvers)}
    latex_dict["numFeatureExtractors"] = str(len(
        [p for p in extractor_path.iterdir() if p.is_dir()]))
    latex_dict["featureExtractorList"] = get_feature_extractor_list(extractor_path)
    latex_dict["numInstanceClasses"] = get_num_instance_sets(train_data.instances)
    latex_dict["instanceClassList"] = get_instance_set_count_list(train_data.instances)
    latex_dict["featureComputationCutoffTime"] = str(extractor_cutoff)
    latex_dict["performanceComputationCutoffTime"] = str(cutoff)
    rank_list_perfect = scmch.compute_perfect_selector_marginal_contribution(train_data)
    rank_list_actual = scmch.compute_actual_selector_marginal_contribution(
        train_data, feature_data, performance_cutoff=cutoff)
    latex_dict["solverPerfectRankingList"] = solver_rank_list_latex(rank_list_perfect)
    latex_dict["solverActualRankingList"] = solver_rank_list_latex(rank_list_actual)
    latex_dict["PARRankingList"] = get_par_ranking_list(train_data, objective)
    latex_dict["VBSPAR"] = str(train_data.calc_vbs_penalty_time())
    latex_dict["actualPAR"] = get_actual_par(actual_performance_dict)
    latex_dict["metric"] = objective.metric
    latex_dict["figure-portfolio-selector-sparkle-vs-sbs"] =\
        get_figure_portfolio_selector_sparkle_vs_sbs(target_dir, objective, train_data,
                                                     actual_performance_dict, penalty)
    latex_dict["figure-portfolio-selector-sparkle-vs-vbs"] =\
        get_figure_portfolio_selector_sparkle_vs_vbs(target_dir, objective, train_data,
                                                     actual_performance_dict, penalty)
    latex_dict["testBool"] = r"\testfalse"

    # Train and test
    if test_case_data is not None:
        latex_dict["testInstanceClass"] =\
            f"\\textbf{ {test_case_data.csv_filepath.parent.name} }"
        latex_dict["numInstanceInTestInstanceClass"] =\
            str(test_case_data.num_instances)
        latex_dict["testActualPAR"] = get_test_actual_par(test_case_data)
        latex_dict["testBool"] = r"\testtrue"

    return latex_dict


def fill_template_tex(template_tex: str, variables: dict) -> str:
    """Given a latex template, replaces all the @@ variables using the dict.

    Args:
        template_tex: The template to be populated
        variables: Variable names (key) with their target (value)

    Returns:
        The populated latex string.
    """
    for variable_key, target_value in variables.items():
        variable = "@@" + variable_key + "@@"
        # We don't modify variable names in the Latex file
        if "\\includegraphics" not in target_value and "\\label" not in target_value:
            # Rectify underscores in target_value
            target_value = target_value.replace("_", r"\textunderscore ")
        template_tex = template_tex.replace(variable, target_value)
    return template_tex


def generate_report(latex_source_path: Path,
                    latex_template_name: str,
                    target_path: Path,
                    report_name: str,
                    variable_dict: dict) -> None:
    """General steps to generate a report.

    Args:
        latex_source_path: The path to the template
        latex_template_name: The template name
        target_path: The directory where the result should be placed
        report_name: The name of the pdf (without suffix)
        variable_dict: TBD
    """
    latex_template_filepath = latex_source_path / latex_template_name

    report_content = latex_template_filepath.open("r").read()
    report_content = fill_template_tex(report_content, variable_dict)

    target_path.mkdir(parents=True, exist_ok=True)
    latex_report_filepath = target_path / report_name
    latex_report_filepath = latex_report_filepath.with_suffix(".tex")
    Path(latex_report_filepath).open("w+").write(report_content)

    stex.check_tex_commands_exist(target_path)
    report_path = stex.compile_pdf(target_path, report_name)

    print(f"Report is placed at: {report_path}")


def generate_gnuplot(output_gnuplot_script: str,
                     output_dir: Path = None) -> None:
    """Generates plot from script using GNU plot."""
    subprocess_plot = subprocess.run(["gnuplot", output_gnuplot_script],
                                     capture_output=True,
                                     cwd=output_dir)

    if subprocess_plot.returncode != 0:
        print(f"(GnuPlot) Error whilst plotting {output_gnuplot_script}:\n"
              f"{subprocess_plot.stderr.decode()}\n")


def generate_pdf(eps_file: str,
                 output_dir: Path = None) -> None:
    """Generate PDF using epstopdf."""
    # Some systems are missing epstopdf so a copy is included
    epsbackup = Path(__file__).parent.parent.resolve() / "Components/epstopdf.pl"
    epstopdf = which("epstopdf") or epsbackup
    subprocess_epstopdf = subprocess.run([epstopdf, eps_file],
                                         capture_output=True,
                                         cwd=output_dir)

    if subprocess_epstopdf.returncode != 0:
        print(f"(Eps To PDF) Error whilst converting Eps to PDF {eps_file}"
              f"{subprocess_epstopdf.stderr.decode()}")


def generate_comparison_plot(points: list,
                             figure_filename: str,
                             xlabel: str = "default",
                             ylabel: str = "optimised",
                             title: str = "",
                             scale: str = "log",
                             limit: str = "magnitude",
                             limit_min: float = 0.2,
                             limit_max: float = 0.2,
                             penalty_time: float = None,
                             replace_zeros: bool = True,
                             magnitude_lines: int = 2147483647,
                             output_dir: Path = None) -> None:
    """Create comparison plots between two different solvers/portfolios.

    Args:
        points: list of points which represents with the performance results of
        (solverA, solverB)
        figure_filename: filename without filetype (e.g., .jpg) to save the figure to.
        xlabel: Name of solverA (default: default)
        ylabel: Name of solverB (default: optimised)
        title: Display title in the image (default: None)
        scale: [linear, log] (default: linear)
        limit: The method to compute the axis limits in the figure
            [absolute, relative, magnitude] (default: relative)
            absolute: Uses the limit_min/max values as absolute values
            relative: Decreases/increases relatively to the min/max values found in the
            points. E.g., min/limit_min and max*limit_max
            magnitude: Increases the order of magnitude(10) of the min/max values in the
            points. E.g., 10**floor(log10(min)-limit_min)
            and 10**ceil(log10(max)+limit_max)
        limit_min: Value used to compute the minimum limit
        limit_max: Value used to compute the maximum limit
        penalty_time: Acts as the maximum value the figure takes in consideration for
        computing the figure limits. This is only relevant for runtime objectives
        replace_zeros: Replaces zeros valued performances to a very small value to make
        plotting on log-scale possible
        magnitude_lines: Draw magnitude lines (only supported for log scale)
        output_dir: directory path to place the figure and its intermediate files in
            (default: current working directory)
    """
    if output_dir is None:
        output_dir = Path()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    points = np.array(points)
    if replace_zeros:
        zero_runtime = 0.000001  # Microsecond
        if np.any(points <= 0):
            print("WARNING: Zero or negative valued performance values detected. Setting"
                  f" these values to {zero_runtime}.")
        points[points <= 0] = zero_runtime

    # process labels
    # LaTeX safe formatting
    xlabel = xlabel.replace("_", "\\_").replace("$", "\\$").replace("^", "\\^")
    ylabel = ylabel.replace("_", "\\_").replace("$", "\\$").replace("^", "\\^")

    # process range values
    min_point_value = np.min(points)
    max_point_value = np.max(points)
    if penalty_time is not None:
        if (penalty_time < max_point_value):
            print("ERROR: Penalty time too small for the given performance data.")
            sys.exit(-1)
        max_point_value = penalty_time

    if limit == "absolute":
        min_value = limit_min
        max_value = limit_max
    elif limit == "relative":
        min_value = (min_point_value * (1 / limit_min) if min_point_value > 0
                     else min_point_value * limit_min)
        max_value = (max_point_value * limit_max if max_point_value > 0
                     else max_point_value * (1 / limit_max))
    elif limit == "magnitude":
        min_value = 10 ** (np.floor(np.log10(min_point_value)) - limit_min)
        max_value = 10 ** (np.ceil(np.log10(max_point_value)) + limit_max)

    if scale == "log" and np.min(points) <= 0:
        raise Exception("Cannot plot negative and zero values on a log scales")

    output_data_file = f"{figure_filename}.dat"
    output_gnuplot_script = output_dir / f"{figure_filename}.plt"
    output_eps_file = f"{figure_filename}.eps"

    # Create data file
    with (output_dir / output_data_file).open("w") as fout:
        for point in points:
            fout.write(" ".join([str(c) for c in point]) + "\n")

    # Generate plot script
    with output_gnuplot_script.open("w") as fout:
        fout.write(f"set xlabel '{xlabel}'\n"
                   f"set ylabel '{ylabel}'\n"
                   f"set title '{title}'\n"
                   "unset key\n"
                   f"set xrange [{min_value}:{max_value}]\n"
                   f"set yrange [{min_value}:{max_value}]\n")
        if scale == "log":
            fout.write("set logscale x\n"
                       "set logscale y\n")
        fout.write("set grid lc rgb '#CCCCCC' lw 2\n"
                   "set size square\n"
                   f"set arrow from {min_value},{min_value} to {max_value},{max_value}"
                   " nohead lc rgb '#AAAAAA'\n")
        # TODO magnitude lines for linear scale
        if magnitude_lines > 0 and scale == "log":
            for order in range(1, magnitude_lines + 1):
                min_shift = min_value * 10 ** order
                max_shift = 10**(np.log10(max_value) - order)
                if min_shift >= max_value:  # Outside plot
                    # Only print magnitude lines if the fall within the visible plotting
                    # area.
                    break

                fout.write(f"set arrow from {min_value},{min_shift} to {max_shift},"
                           f"{max_value} nohead lc rgb '#CCCCCC' dashtype '-'\n"
                           f"set arrow from {min_shift},{min_value} to {max_value},"
                           f"{max_shift} nohead lc rgb '#CCCCCC' dashtype '-'\n")

        if penalty_time is not None:
            fout.write(f"set arrow from {min_value},{penalty_time} to {max_value},"
                       f"{penalty_time} nohead lc rgb '#AAAAAA'\n"
                       f"set arrow from {penalty_time},{min_value} to {penalty_time},"
                       f"{max_value} nohead lc rgb '#AAAAAA'\n")

        fout.write('set terminal postscript eps color solid linewidth "Helvetica" 20\n'
                   f"set output '{output_eps_file}\n"
                   "set style line 1 pt 2 ps 1.5 lc rgb 'royalblue' \n"
                   f"plot '{output_data_file}' ls 1\n")

    generate_gnuplot(output_gnuplot_script.name, output_dir)
    generate_pdf(output_eps_file, output_dir)
    output_gnuplot_script.unlink(missing_ok=True)


def generate_report_selection(target_path: Path,
                              latex_dir: Path,
                              latex_template: Path,
                              bibliography_path: Path,
                              extractor_path: Path,
                              selector_path: Path,
                              feature_data: FeatureDataFrame,
                              train_data: PerformanceDataFrame,
                              extractor_cutoff: int,
                              cutoff: int,
                              penalty: int,
                              test_case_data: PerformanceDataFrame = None) -> None:
    """Generate a report for algorithm selection.

    Args:
        target_path: Path where the outputfiles will be placed.
        latex_dir: The latex dir
        latex_template: The template for the report
        bibliography_path: Path to the bib file.
        extractor_path: Path to the extractor used
        selector_path: Path to the selector
        feature_data: Feature data created by extractor
        train_data: The performance input data for the selector
        extractor_cutoff: The maximum time for the selector to run
        cutoff: The cutoff per solver
        penalty: The penalty for solvers TIMEOUT
        test_case_data: Path to the test case directory. Defaults to None.
    """
    # Include results on the test set if a test case directory is given
    latex_report_filename = Path("Sparkle_Report")
    if test_case_data is not None:
        latex_report_filename = Path("Sparkle_Report_for_Test")

    target_path.mkdir(parents=True, exist_ok=True)
    dict_variable_to_value = selection_report_variables(target_path,
                                                        bibliography_path,
                                                        extractor_path,
                                                        selector_path,
                                                        feature_data,
                                                        extractor_cutoff,
                                                        cutoff,
                                                        penalty,
                                                        train_data,
                                                        test_case_data)
    generate_report(latex_dir,
                    latex_template,
                    target_path,
                    latex_report_filename,
                    dict_variable_to_value)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for the execution of a (configured) solver on instances."""
from __future__ import annotations

from pathlib import Path

import runrunner as rrr
from runrunner.base import Runner, Run
from runrunner.slurm import SlurmRun

from sparkle.CLI.help import global_variables as gv
from sparkle.platform import CommandName
import sparkle.tools.general as tg
from sparkle.solver import Solver
from sparkle.instance import InstanceSet
from sparkle.tools.runsolver_parsing import get_solver_output


def call_solver(
        instance_set: InstanceSet,
        solver: Solver,
        config: str | Path = None,
        seed: int | list[int] = 42,
        outdir: Path = None,
        commandname: CommandName = CommandName.RUN_SOLVERS,
        dependency: SlurmRun | list[SlurmRun] = None,
        run_on: Runner = Runner.SLURM) -> Run:
    """Run a solver on all given instances.

    Args:
        instance_list: A list of all paths in a directory of instances.
        solver: The solver to run on the instances
        config: The configuration with which to run. Can be direct configuration string,
            or file from which to read. If specific line from file is needed, seed
            should be specified.
        seed: The seed for the solver.
        outdir: Path where to place the output files of the (run) solver logs
        commandname: The commandname under which to run the process.
        dependency: The jobs it depends on to finish before starting.
        run_on: Whether the command is run with Slurm or not.

    Returns:
        The Runrunner Run object regarding the call.
    """
    custom_cutoff = gv.settings().get_general_target_cutoff_time()
    cmd_list = []
    runsolver_args_list = []
    solver_params_list = []
    for index, instance_path in enumerate(instance_set.instance_paths):
        raw_result_path = Path(f"{solver.name}_{instance_set._instance_names[index]}"
                               f"_{tg.get_time_pid_random_string()}.rawres")
        runsolver_watch_data_path = raw_result_path.with_suffix(".log")
        runsolver_values_path = raw_result_path.with_suffix(".val")

        runsolver_args = ["--timestamp", "--use-pty",
                          "--cpu-limit", str(custom_cutoff),
                          "-w", runsolver_watch_data_path,
                          "-v", runsolver_values_path,
                          "-o", raw_result_path]
        if isinstance(config, str):
            solver_params = solver.config_str_to_dict(config)
        elif isinstance(config, Path):
            solver_params = {"config_path": config}
        else:
            solver_params = {}
        solver_params["specifics"] = "rawres"
        solver_params["cutoff_time"] = custom_cutoff
        solver_params["run_length"] = "2147483647"  # Arbitrary, not used by SMAC wrapper
        if seed is None:
            solver_params["seed"] = gv.get_seed()
        else:
            # Use the seed to determine the configuration line in the file
            if isinstance(seed, list):
                solver_params["seed"] = seed[index]
            else:
                solver_params["seed"] = seed
        runsolver_args_list.append(runsolver_args)
        solver_params_list.append(solver_params)
        if isinstance(instance_path, list):
            instance_path = [p.absolute() for p in instance_path]
        else:
            instance_path = instance_path.absolute()
        solver_cmd = solver.build_cmd(instance_path,
                                      solver_params, runsolver_args)
        cmd_list.append(" ".join(solver_cmd))

    sbatch_options = gv.settings().get_slurm_extra_options(as_args=True)
    srun_options = ["-N1", "-n1"] + sbatch_options
    # Make sure the executable dir exists
    if outdir is None:
        outdir = solver.raw_output_directory
    outdir.mkdir(exist_ok=True, parents=True)

    if run_on == Runner.LOCAL:
        print(f"\nStart running solver on {instance_set.size} instances...")
    run = rrr.add_to_queue(
        runner=run_on,
        cmd=cmd_list,
        name=commandname,
        base_dir=gv.settings().DEFAULT_tmp_output,
        path=outdir,
        dependencies=dependency,
        sbatch_options=sbatch_options,
        srun_options=srun_options)

    if run_on == Runner.LOCAL:
        # Wait for all jobs to complete before printing
        run.wait()
        # Jobs are sorted in the cmd list order
        for index, job in enumerate(run.jobs):
            # Run the configured solver
            job.wait()
            output_log = outdir / runsolver_args_list[index][-1]
            solver_output = get_solver_output(runsolver_args_list[index],
                                              output_log.read_text(),
                                              solver.raw_output_directory)
            # Output results to user, including path to rawres_solver
            print(f"Execution of {solver.name} on instance {Path(instance_path).name} "
                  f"completed with status {solver_output['status']} in "
                  f"{solver_output['runtime']} seconds.")
            print("Raw output can be found at: "
                  f"{solver.raw_output_directory / raw_result_path}.\n")

    return run



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for job execution and maintenance."""
from __future__ import annotations

from pathlib import Path
import time
import json

from runrunner.slurm import SlurmRun
from runrunner.base import Status
from tabulate import tabulate

from sparkle.CLI.help import global_variables as gv
from sparkle.platform.cli_types import VerbosityLevel, TEXT


def wait_for_job(job: str | SlurmRun, path: Path) -> None:
    """Wait for a given Slurm job to finish executing.

    Args:
      job: String job identifier.
      path: The Path where to look for the stored jobs.
    """
    if isinstance(job, str):
        job = find_run(job, path)
    job.wait()

    print(f"Job with ID {job.run_id} done!", flush=True)


def find_run(job_id: str, path: Path) -> SlurmRun:
    """Retrieve a specific RunRunner Slurm run using the job_id.

    Args:
        job_id: String identifier of the job
        path: The Path where to look for the stored jobs.

    Returns:
        The SlurmRun with the matching run_id, None if no match.
    """
    for run in get_runs_from_file(path=path):
        if run.run_id == job_id:
            return run
    return None


def get_runs_from_file(path: Path) -> list[SlurmRun]:
    """Retrieve all run objects from file storage.

    Args:
        path: Path object where to look for the files.

    Returns:
        List of all found SlumRun objects.
    """
    runs, ignores = [], []
    if not path.exists():
        return runs
    for file in path.iterdir():
        if file.suffix == ".json":
            # TODO: RunRunner should be adapted to have more general methods for runs
            # So this method can work for both local and slurm
            try:
                run_obj = SlurmRun.from_file(file)
                runs.append(run_obj)
            except Exception:
                # TODO: When we have a more sophisticated version of managing/
                # remembering runs, we might as well remember which files to ignore.
                ignores.append(file)
    return runs


def get_running_jobs(path: Path) -> list[SlurmRun]:
    """Returns all waiting or running jobs."""
    return [run for run in get_runs_from_file(path)
            if run.status == Status.WAITING or run.status == Status.RUNNING]


def get_dependencies(jobs: list[SlurmRun]) -> list[SlurmRun]:
    """Adds Dependencies. Should be removed once RunRunner provides this feature."""
    for job in jobs:
        with job.json_filepath.open() as file:
            data = json.load(file)
        if len(data["dependencies"]) > 0:
            job_ids = [dep["run_id"] for dep in data["dependencies"]]
        else:
            job_ids = []
        job.dependencies = job_ids
    return jobs


def clear_console_lines(lines: int) -> None:
    """Clears the last lines of the console."""
    # \033 is the escape character (ESC) in ASCII
    # [{lines}A is the escape sequence that moves the cursor up.
    print(f"\033[{lines}A", end="")
    # [J is the exape sequence that clears the console from the cursor down
    print("\033[J", end="")


def wait_for_all_jobs(path: Path) -> None:
    """Wait for all active jobs to finish executing."""
    jobs = get_running_jobs(path)
    verbosity_setting = gv.settings().get_general_verbosity()
    running_jobs = jobs
    # Interval at which to refresh the table
    check_interval = gv.settings().get_general_check_interval()
    # If verbosity is quiet there is no need for further information
    if verbosity_setting == VerbosityLevel.QUIET:
        prev_jobs = len(running_jobs) + 1
        while len(running_jobs) > 0:
            if len(running_jobs) < prev_jobs:
                print(f"Waiting for {len(running_jobs)} jobs...", flush=True)
            time.sleep(check_interval)
            prev_jobs = len(running_jobs)
            running_jobs = [run for run in running_jobs
                            if run.status == Status.WAITING
                            or run.status == Status.RUNNING]

    # If verbosity is standard the command will print a table with relevant information
    elif verbosity_setting == VerbosityLevel.STANDARD:
        # Collect dependencies and partitions for each job
        jobs = get_dependencies(jobs)
        # Order in which to display the jobs
        status_order = {Status.COMPLETED: 0, Status.RUNNING: 1, Status.WAITING: 2}
        while len(running_jobs) > 0:
            # Information to be printed to the table
            information = [["RunId", "Name", "Partition", "Status",
                            "Dependencies", "Finished Jobs", "Run Time"]]
            running_jobs = [run for run in running_jobs
                            if run.status == Status.WAITING
                            or run.status == Status.RUNNING]
            sorted_jobs = sorted(
                jobs, key=lambda job: (status_order.get(job.status, 4), job.run_id))
            for job in sorted_jobs:
                # Count number of jobs that have finished
                finished_jobs_count = sum(1 for status in job.all_status
                                          if status == Status.COMPLETED)
                # Format job.status
                status_text = \
                    TEXT.format_text([TEXT.BOLD], job.status) \
                    if job.status == Status.RUNNING else \
                    (TEXT.format_text([TEXT.ITALIC], job.status)
                        if job.status == Status.COMPLETED else job.status)
                information.append(
                    [job.run_id,
                     job.name,
                     job.partition,
                     status_text,
                     "None" if len(job.dependencies) == 0
                        else ", ".join(job.dependencies),
                     f"{finished_jobs_count}/{len(job.all_status)}",
                     job.runtime])
            # Print the table
            table = tabulate(information, headers="firstrow", tablefmt="grid")
            print(table)
            time.sleep(check_interval)

            # Clears the table for the new table to be printed
            lines = table.count("\n") + 1
            clear_console_lines(lines)

    print("All jobs done!")
